from __future__ import annotations

from pathlib import Path
from typing import Any
from urllib.parse import urlparse

import aiosonic
from llm_async.utils.retry import RetryConfig

from minibot.adapters.config.schema import LLMMConfig
from minibot.llm.services.provider_registry import resolve_provider_class


def create_provider(config: LLMMConfig) -> tuple[Any, str]:
    configured_provider = config.provider.lower()
    provider_cls, provider_name = resolve_provider_class(configured_provider)
    normalized_base_url = config.base_url.rstrip("/") if config.base_url else None

    timeouts = aiosonic.Timeouts(
        sock_connect=float(config.sock_connect_timeout_seconds),
        sock_read=float(config.sock_read_timeout_seconds),
        request_timeout=float(config.request_timeout_seconds),
    )
    connector = aiosonic.TCPConnector(timeouts=timeouts)
    retry_config = RetryConfig(
        max_attempts=config.retry_attempts + 1,
        base_delay=float(config.retry_delay_seconds),
        max_delay=float(config.retry_delay_seconds),
        backoff_factor=1.0,
        jitter=False,
    )
    effective_http2 = config.http2
    if normalized_base_url:
        parsed_base_url = urlparse(normalized_base_url)
        if parsed_base_url.scheme.lower() == "http":
            effective_http2 = False
    provider_kwargs: dict[str, Any] = {
        "api_key": config.api_key,
        "retry_config": retry_config,
        "client_kwargs": {"connector": connector},
        "http2": effective_http2,
    }
    if normalized_base_url:
        provider_kwargs["base_url"] = normalized_base_url

    return provider_cls(**provider_kwargs), provider_name


def load_system_prompt(config: LLMMConfig) -> str:
    prompt_file = getattr(config, "system_prompt_file", None)
    if prompt_file is not None:
        normalized = prompt_file.strip() if isinstance(prompt_file, str) else None
        if normalized:
            prompt_path = Path(normalized)
            if not prompt_path.exists():
                raise FileNotFoundError(f"system_prompt_file configured but file not found: {normalized}")
            if not prompt_path.is_file():
                raise ValueError(f"system_prompt_file configured but path is not a file: {normalized}")
            content = prompt_path.read_text(encoding="utf-8").strip()
            if not content:
                raise ValueError(f"system_prompt_file configured but file is empty: {normalized}")
            return content
    return getattr(config, "system_prompt", "You are Minibot, a helpful assistant.")


def build_openrouter_provider_payload(config: LLMMConfig) -> dict[str, Any]:
    provider_cfg = getattr(getattr(config, "openrouter", None), "provider", None)
    if provider_cfg is None:
        return {}

    payload: dict[str, Any] = dict(getattr(provider_cfg, "provider_extra", {}) or {})
    typed_fields = (
        "order",
        "allow_fallbacks",
        "require_parameters",
        "data_collection",
        "zdr",
        "enforce_distillable_text",
        "only",
        "ignore",
        "quantizations",
        "sort",
        "preferred_min_throughput",
        "preferred_max_latency",
        "max_price",
    )
    for field_name in typed_fields:
        value = getattr(provider_cfg, field_name, None)
        if value is not None:
            payload[field_name] = value
    return payload


def resolve_openrouter_reasoning_enabled(config: LLMMConfig) -> bool | None:
    openrouter_cfg = getattr(config, "openrouter", None)
    if openrouter_cfg is None:
        return None
    value = getattr(openrouter_cfg, "reasoning_enabled", None)
    if isinstance(value, bool):
        return value
    return None



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Sequence


@dataclass(frozen=True)
class RequestContext:
    model: str
    provider_name: str
    is_responses_provider: bool
    temperature: float | None
    max_new_tokens: int | None
    prompt_cache_enabled: bool
    prompt_cache_retention: str | None
    reasoning_effort: str | None
    openrouter_models: tuple[str, ...]
    openrouter_provider: dict[str, Any]
    openrouter_reasoning_enabled: bool | None
    openrouter_plugins: tuple[dict[str, Any], ...]


def build_messages(
    *,
    history: Sequence[Any],
    user_message: str,
    user_content: str | list[dict[str, Any]] | None,
    system_prompt: str,
) -> list[dict[str, Any]]:
    messages = [
        {"role": "system", "content": system_prompt},
    ]
    messages.extend({"role": entry.role, "content": entry.content} for entry in history)
    final_user_content: str | list[dict[str, Any]] = user_message
    if user_content is not None:
        final_user_content = user_content
    messages.append({"role": "user", "content": final_user_content})
    return messages


def build_generate_extra_kwargs(
    *,
    ctx: RequestContext,
    prompt_cache_key: str | None,
    previous_response_id: str | None,
    system_prompt: str,
) -> dict[str, Any]:
    extra_kwargs: dict[str, Any] = {}
    if prompt_cache_key and ctx.is_responses_provider and ctx.prompt_cache_enabled:
        extra_kwargs["prompt_cache_key"] = prompt_cache_key
    if previous_response_id and ctx.is_responses_provider:
        extra_kwargs["previous_response_id"] = previous_response_id
    if ctx.is_responses_provider and ctx.prompt_cache_enabled and ctx.prompt_cache_retention:
        extra_kwargs["prompt_cache_retention"] = ctx.prompt_cache_retention
    if ctx.is_responses_provider and ctx.reasoning_effort:
        extra_kwargs.setdefault("reasoning", {"effort": ctx.reasoning_effort})
    if ctx.is_responses_provider:
        extra_kwargs["instructions"] = system_prompt
    return extra_kwargs


def build_generate_step_call_kwargs(
    *,
    ctx: RequestContext,
    conversation: Sequence[dict[str, Any]],
    tool_specs: Sequence[Any] | None,
    strict_response_schema: dict[str, Any] | None,
    extra_kwargs: dict[str, Any],
) -> dict[str, Any]:
    call_kwargs: dict[str, Any] = {
        "model": ctx.model,
        "messages": list(conversation),
        "tools": tool_specs,
        "response_schema": strict_response_schema,
    }
    if ctx.temperature is not None:
        call_kwargs["temperature"] = ctx.temperature
    if ctx.is_responses_provider:
        if ctx.max_new_tokens is not None:
            call_kwargs["max_output_tokens"] = ctx.max_new_tokens
    else:
        resolved_max_tokens = resolved_max_tokens_for_request(ctx)
        if resolved_max_tokens is not None:
            call_kwargs["max_tokens"] = resolved_max_tokens
    call_kwargs.update(openrouter_kwargs(ctx))
    call_kwargs.update(extra_kwargs)
    return call_kwargs


def build_complete_once_call_kwargs(
    *,
    ctx: RequestContext,
    messages: Sequence[dict[str, Any]],
    tool_specs: Sequence[Any] | None,
    strict_response_schema: dict[str, Any] | None,
    prompt_cache_key: str | None,
    previous_response_id: str | None,
) -> dict[str, Any]:
    call_kwargs: dict[str, Any] = {
        "model": ctx.model,
        "messages": list(messages),
        "tools": tool_specs,
        "response_schema": strict_response_schema,
    }
    if ctx.temperature is not None:
        call_kwargs["temperature"] = ctx.temperature
    if ctx.is_responses_provider:
        if ctx.max_new_tokens is not None:
            call_kwargs["max_output_tokens"] = ctx.max_new_tokens
        instructions = extract_system_instructions(messages)
        if instructions:
            call_kwargs["instructions"] = instructions
    else:
        resolved_max_tokens = resolved_max_tokens_for_request(ctx)
        if resolved_max_tokens is not None:
            call_kwargs["max_tokens"] = resolved_max_tokens
    call_kwargs.update(openrouter_kwargs(ctx))
    if prompt_cache_key and ctx.is_responses_provider and ctx.prompt_cache_enabled:
        call_kwargs["prompt_cache_key"] = prompt_cache_key
    if previous_response_id and ctx.is_responses_provider:
        call_kwargs["previous_response_id"] = previous_response_id
    if ctx.is_responses_provider and ctx.prompt_cache_enabled and ctx.prompt_cache_retention:
        call_kwargs["prompt_cache_retention"] = ctx.prompt_cache_retention
    if ctx.is_responses_provider and ctx.reasoning_effort:
        call_kwargs.setdefault("reasoning", {"effort": ctx.reasoning_effort})
    return call_kwargs


def build_continue_call_kwargs(
    *,
    ctx: RequestContext,
    previous_response_id: str,
    prompt_cache_key: str | None,
    system_prompt: str,
    response_schema: dict[str, Any] | None,
) -> dict[str, Any]:
    call_kwargs: dict[str, Any] = {
        "model": ctx.model,
        "messages": [{"role": "user", "content": "Continue exactly where you left off. Do not repeat."}],
        "tools": None,
        "response_schema": response_schema,
        "previous_response_id": previous_response_id,
        "instructions": system_prompt,
    }
    if ctx.temperature is not None:
        call_kwargs["temperature"] = ctx.temperature
    if ctx.max_new_tokens is not None:
        call_kwargs["max_output_tokens"] = ctx.max_new_tokens
    if prompt_cache_key and ctx.prompt_cache_enabled:
        call_kwargs["prompt_cache_key"] = prompt_cache_key
    if ctx.prompt_cache_enabled and ctx.prompt_cache_retention:
        call_kwargs["prompt_cache_retention"] = ctx.prompt_cache_retention
    if ctx.reasoning_effort:
        call_kwargs.setdefault("reasoning", {"effort": ctx.reasoning_effort})
    return call_kwargs


def extract_system_instructions(messages: Sequence[dict[str, Any]]) -> str | None:
    for message in messages:
        if not isinstance(message, dict):
            continue
        if message.get("role") != "system":
            continue
        content = message.get("content")
        if isinstance(content, str) and content.strip():
            return content
    return None


def openrouter_kwargs(ctx: RequestContext) -> dict[str, Any]:
    if ctx.provider_name != "openrouter":
        return {}
    kwargs: dict[str, Any] = {}
    if ctx.openrouter_models:
        kwargs["models"] = list(ctx.openrouter_models)
    if ctx.openrouter_provider:
        kwargs["provider"] = dict(ctx.openrouter_provider)
    if ctx.openrouter_reasoning_enabled is True:
        kwargs["reasoning"] = {"enabled": True}
    if ctx.openrouter_plugins:
        kwargs["plugins"] = list(ctx.openrouter_plugins)
    return kwargs


def resolved_max_tokens_for_request(ctx: RequestContext) -> int | None:
    if ctx.is_responses_provider:
        return None
    if ctx.provider_name == "openrouter":
        if ctx.max_new_tokens is None:
            return 4096
        return min(ctx.max_new_tokens, 32768)
    if ctx.max_new_tokens is not None:
        return ctx.max_new_tokens
    return None



from __future__ import annotations

from dataclasses import dataclass
from typing import Any

from minibot.core.agent_runtime import AgentState


@dataclass(frozen=True)
class DelegationTraceResult:
    trace: list[dict[str, Any]]
    fallback_used: bool
    unresolved: bool


def count_tool_messages(state: AgentState) -> int:
    return sum(1 for message in state.messages if message.role == "tool")


def extract_delegation_trace(state: AgentState) -> DelegationTraceResult:
    trace: list[dict[str, Any]] = []
    fallback_used = False
    last_result_status: str | None = None
    last_should_answer: bool | None = None
    saw_invoke = False
    for message in state.messages:
        if message.role != "tool" or message.name != "invoke_agent":
            continue
        if not message.content:
            continue
        part = message.content[0]
        if part.type != "json" or not isinstance(part.value, dict):
            continue
        saw_invoke = True
        agent = str(part.value.get("agent") or "")
        ok = bool(part.value.get("ok", False))
        error = part.value.get("error")
        result_status = part.value.get("result_status")
        delegated_should_answer = part.value.get("should_answer_to_user")
        if not ok:
            fallback_used = True
        if isinstance(result_status, str):
            last_result_status = result_status
        else:
            last_result_status = None
        if isinstance(delegated_should_answer, bool):
            last_should_answer = delegated_should_answer
        else:
            last_should_answer = None
        trace_entry: dict[str, Any] = {
            "agent": "minibot",
            "decision": "invoke_agent",
            "target": agent or None,
            "ok": ok,
        }
        if isinstance(result_status, str) and result_status.strip():
            trace_entry["result_status"] = result_status
        if isinstance(delegated_should_answer, bool):
            trace_entry["should_answer_to_user"] = delegated_should_answer
        if isinstance(error, str) and error.strip():
            trace_entry["error"] = error
        trace.append(trace_entry)

    unresolved = False
    if saw_invoke:
        if last_result_status in {"invalid_result", "not_user_answerable"}:
            unresolved = True
        if last_should_answer is False:
            unresolved = True

    return DelegationTraceResult(
        trace=trace,
        fallback_used=fallback_used,
        unresolved=unresolved,
    )



from __future__ import annotations

from dataclasses import dataclass
import logging
from typing import Any, Sequence

from minibot.app.agent_runtime import AgentRuntime
from minibot.app.incoming_files_context import (
    build_history_user_entry,
)
from minibot.app.handlers.services import (
    AgentRuntimeResult,
    CompactionResult,
    HistoryCompactionService,
    PromptService,
    ResponseMetadataService,
    RuntimeOrchestrationService,
    SessionStateService,
    UserInputService,
)
from minibot.app.runtime_limits import build_runtime_limits
from minibot.app.response_parser import extract_answer, plain_render
from minibot.app.tool_use_guardrail import NoopToolUseGuardrail, ToolUseGuardrail
from minibot.core.channels import ChannelMessage, ChannelResponse
from minibot.core.events import MessageEvent
from minibot.core.memory import MemoryBackend
from minibot.llm.provider_factory import LLMClient
from minibot.shared.assistant_response import assistant_response_schema
from minibot.shared.utils import session_id_for, session_id_from_parts
from minibot.llm.tools.base import ToolBinding, ToolContext


@dataclass(frozen=True)
class _CompactionResult(CompactionResult):
    updates: list[str]
    performed: bool
    tokens_used: int
    session_total_tokens_before_compaction: int | None
    session_total_tokens_after_compaction: int


@dataclass(frozen=True)
class _AgentRuntimeResult(AgentRuntimeResult):
    render: Any
    should_reply: bool
    response_id: str | None
    agent_trace: list[dict[str, Any]]
    delegation_fallback_used: bool
    tokens_used: int


class LLMMessageHandler:
    _COMPACTION_USER_REQUEST = "Please compact the current conversation memory."

    def __init__(
        self,
        memory: MemoryBackend,
        llm_client: LLMClient,
        tools: Sequence[ToolBinding] | None = None,
        default_owner_id: str | None = None,
        max_history_messages: int | None = None,
        max_history_tokens: int | None = None,
        notify_compaction_updates: bool = False,
        agent_timeout_seconds: int = 120,
        environment_prompt_fragment: str = "",
        tool_use_guardrail: ToolUseGuardrail | None = None,
        managed_files_root: str | None = None,
    ) -> None:
        self._memory = memory
        self._llm_client = llm_client
        self._tools = list(tools or [])
        self._default_owner_id = default_owner_id
        self._max_history_messages = max_history_messages
        self._max_history_tokens = max_history_tokens
        self._notify_compaction_updates = notify_compaction_updates
        self._tool_use_guardrail: ToolUseGuardrail = tool_use_guardrail or NoopToolUseGuardrail()
        self._logger = logging.getLogger("minibot.handler")
        self._session_state = SessionStateService()
        self._session_total_tokens = self._session_state.session_total_tokens
        self._session_previous_response_ids = self._session_state.session_previous_response_ids
        self._metadata_service = ResponseMetadataService(llm_client=self._llm_client)
        self._input_service = UserInputService(llm_client=self._llm_client)
        self._prompt_service = PromptService(
            llm_client=self._llm_client,
            tools=self._tools,
            environment_prompt_fragment=environment_prompt_fragment,
            logger=self._logger,
        )
        self._prompts_dir = self._prompt_service.prompts_dir
        runtime_limits = build_runtime_limits(
            llm_client=self._llm_client,
            timeout_seconds=agent_timeout_seconds,
            min_timeout_seconds=120,
        )
        if self._supports_agent_runtime():
            self._runtime = AgentRuntime(
                llm_client=self._llm_client,
                tools=self._tools,
                limits=runtime_limits,
                allowed_append_message_tools=["self_insert_artifact"],
                allow_system_inserts=False,
                managed_files_root=managed_files_root,
            )
            self._runtime_service = RuntimeOrchestrationService(
                runtime=self._runtime,
                llm_client=self._llm_client,
                guardrail=self._tool_use_guardrail,
                session_state=self._session_state,
                logger=self._logger,
            )
        else:
            self._runtime = None
            self._runtime_service = None
        self._compaction_service = HistoryCompactionService(
            memory=self._memory,
            llm_client=self._llm_client,
            session_state=self._session_state,
            prompt_service=self._prompt_service,
            logger=self._logger,
            max_history_tokens=self._max_history_tokens,
            compaction_user_request=self._COMPACTION_USER_REQUEST,
        )

    def _supports_agent_runtime(self) -> bool:
        return callable(getattr(self._llm_client, "complete_once", None)) and callable(
            getattr(self._llm_client, "execute_tool_calls_for_runtime", None)
        )

    def _llm_provider_name(self) -> str | None:
        return self._metadata_service.provider_name()

    def _llm_model_name(self) -> str | None:
        return self._metadata_service.model_name()

    def _response_metadata(self, should_reply: bool) -> dict[str, Any]:
        return self._metadata_service.response_metadata(should_reply)

    def _supports_media_inputs(self) -> bool:
        return self._input_service.supports_media_inputs()

    def _media_input_mode(self) -> str:
        return self._input_service.media_input_mode()

    def _llm_prompts_dir(self) -> str:
        return self._prompt_service.prompts_dir

    def _responses_state_mode(self) -> str:
        mode_getter = getattr(self._llm_client, "responses_state_mode", None)
        if callable(mode_getter):
            mode = mode_getter()
            if mode in {"full_messages", "previous_response_id"}:
                return mode
        return "full_messages"

    def _prompt_cache_enabled(self) -> bool:
        enabled_getter = getattr(self._llm_client, "prompt_cache_enabled", None)
        if callable(enabled_getter):
            return bool(enabled_getter())
        return True

    def _compose_system_prompt(self, channel: str | None) -> str:
        return self._prompt_service.compose_system_prompt(channel)

    async def handle(self, event: MessageEvent) -> ChannelResponse:
        message = event.message
        session_id = session_id_for(message)
        turn_total_tokens = 0
        model_text, model_user_content = self._build_model_user_input(message)
        if message.attachments:
            self._logger.debug(
                "prepared multimodal message",
                extra={
                    "channel": message.channel,
                    "chat_id": message.chat_id,
                    "user_id": message.user_id,
                    "attachment_count": len(message.attachments),
                    "attachment_types": [str(attachment.get("type", "unknown")) for attachment in message.attachments],
                    "responses_provider": self._llm_client.is_responses_provider(),
                    "media_input_mode": self._media_input_mode(),
                },
            )
        await self._memory.append_history(session_id, "user", build_history_user_entry(message, model_text))
        await self._enforce_history_limit(session_id)

        if message.attachments and not self._supports_media_inputs():
            answer = "Media inputs are supported for `openai_responses`, `openai`, and `openrouter`."
            await self._memory.append_history(session_id, "assistant", answer)
            await self._enforce_history_limit(session_id)
            chat_id = message.chat_id or message.user_id or 0
            metadata = self._response_metadata(True)
            metadata["token_trace"] = self._build_token_trace(
                turn_total_tokens=0,
                session_total_tokens_before_compaction=None,
                session_total_tokens_after_compaction=self._current_session_tokens(session_id),
                compaction_performed=False,
            )
            return ChannelResponse(
                channel=message.channel,
                chat_id=chat_id,
                text=answer,
                render=plain_render(answer),
                metadata=metadata,
            )

        history = list(await self._memory.get_history(session_id))
        owner_id = resolve_owner_id(message, self._default_owner_id)
        responses_state_mode = self._responses_state_mode()
        use_previous_response_id = (
            self._llm_client.is_responses_provider() and responses_state_mode == "previous_response_id"
        )
        previous_response_id = (
            self._session_previous_response_ids.get(session_id) if use_previous_response_id else None
        )
        tool_context = ToolContext(
            owner_id=owner_id,
            channel=message.channel,
            chat_id=message.chat_id,
            user_id=message.user_id,
        )
        prompt_cache_key = _prompt_cache_key(message) if self._prompt_cache_enabled() else None
        system_prompt = self._compose_system_prompt(message.channel)
        primary_agent = "minibot"
        agent_trace: list[dict[str, Any]] = []
        delegation_fallback_used = False
        try:
            if self._runtime is None:
                generation = await self._llm_client.generate(
                    history,
                    model_text,
                    user_content=model_user_content,
                    tools=self._tools,
                    tool_context=tool_context,
                    response_schema=self._response_schema(),
                    prompt_cache_key=prompt_cache_key,
                    previous_response_id=previous_response_id,
                    system_prompt_override=system_prompt,
                )
                self._session_state.track_usage(
                    session_id,
                    input_tokens=getattr(generation, "input_tokens", None),
                    output_tokens=getattr(generation, "output_tokens", None),
                    total_tokens=getattr(generation, "total_tokens", None),
                    cached_input_tokens=getattr(generation, "cached_input_tokens", None),
                    reasoning_output_tokens=getattr(generation, "reasoning_output_tokens", None),
                )
                turn_total_tokens += self._track_token_usage(session_id, getattr(generation, "total_tokens", None))
                render, should_reply = extract_answer(generation.payload, logger=self._logger)
                if use_previous_response_id and generation.response_id:
                    self._session_previous_response_ids[session_id] = generation.response_id
            else:
                runtime_result = await self._run_with_agent_runtime(
                    session_id=session_id,
                    history=history,
                    model_text=model_text,
                    model_user_content=model_user_content,
                    system_prompt=system_prompt,
                    tool_context=tool_context,
                    prompt_cache_key=prompt_cache_key,
                    previous_response_id=previous_response_id,
                    chat_id=message.chat_id,
                    channel=message.channel,
                )
                turn_total_tokens += runtime_result.tokens_used
                render = runtime_result.render
                should_reply = runtime_result.should_reply
                agent_trace = runtime_result.agent_trace
                delegation_fallback_used = runtime_result.delegation_fallback_used
                if use_previous_response_id and runtime_result.response_id:
                    self._session_previous_response_ids[session_id] = runtime_result.response_id

            if not use_previous_response_id:
                self._session_previous_response_ids.pop(session_id, None)

            self._logger.debug(
                "structured output parsed",
                extra={"kind": render.kind, "content_length": len(render.text), "should_reply": should_reply},
            )
        except Exception as exc:
            self._logger.exception("LLM call failed", exc_info=exc)
            render = plain_render(self._format_runtime_error_message(exc))
            should_reply = True
        answer = render.text
        await self._memory.append_history(session_id, "assistant", answer)
        await self._enforce_history_limit(session_id)
        compact_prompt_cache_key = prompt_cache_key or f"{session_id}:runtime"
        compaction_result = await self._compact_history_if_needed(
            session_id,
            prompt_cache_key=compact_prompt_cache_key,
            system_prompt=system_prompt,
            notify=self._notify_compaction_updates,
        )
        turn_total_tokens += compaction_result.tokens_used

        chat_id = message.chat_id or message.user_id or 0
        metadata = self._response_metadata(should_reply)
        metadata["primary_agent"] = primary_agent
        if agent_trace:
            metadata["agent_trace"] = agent_trace
        metadata["delegation_fallback_used"] = delegation_fallback_used
        if compaction_result.updates:
            metadata["compaction_updates"] = compaction_result.updates
        metadata["token_trace"] = self._build_token_trace(
            turn_total_tokens=turn_total_tokens,
            session_total_tokens_before_compaction=compaction_result.session_total_tokens_before_compaction,
            session_total_tokens_after_compaction=compaction_result.session_total_tokens_after_compaction,
            compaction_performed=compaction_result.performed,
        )
        usage_trace = self._session_state.latest_usage_trace(session_id)
        if any(value is not None for value in usage_trace.values()):
            metadata["usage_trace"] = usage_trace
        return ChannelResponse(
            channel=message.channel,
            chat_id=chat_id,
            text=answer,
            render=render,
            metadata=metadata,
        )

    async def repair_format_response(
        self,
        *,
        response: ChannelResponse,
        parse_error: str,
        channel: str,
        chat_id: int,
        user_id: int | None,
        attempt: int,
    ) -> ChannelResponse:
        session_id = session_id_from_parts(channel, chat_id, user_id)
        turn_total_tokens = 0
        history = list(await self._memory.get_history(session_id))
        system_prompt = self._compose_system_prompt(channel)
        responses_state_mode = self._responses_state_mode()
        use_previous_response_id = (
            self._llm_client.is_responses_provider() and responses_state_mode == "previous_response_id"
        )
        previous_response_id = (
            self._session_previous_response_ids.get(session_id) if use_previous_response_id else None
        )
        original_kind = response.render.kind if response.render is not None else "text"
        original_content = response.render.text if response.render is not None else response.text
        repair_prompt = self._build_format_repair_prompt(
            channel=channel,
            original_kind=original_kind,
            parse_error=parse_error,
            original_content=original_content,
        )
        generation = await self._llm_client.generate(
            history,
            repair_prompt,
            user_content=None,
            tools=[],
            tool_context=None,
            response_schema=self._response_schema(),
            prompt_cache_key=f"{channel}:{chat_id}:format-repair",
            previous_response_id=previous_response_id,
            system_prompt_override=system_prompt,
        )
        if use_previous_response_id and generation.response_id:
            self._session_previous_response_ids[session_id] = generation.response_id
        turn_total_tokens += self._track_token_usage(session_id, getattr(generation, "total_tokens", None))
        render, _ = extract_answer(generation.payload, logger=self._logger)
        await self._memory.append_history(session_id, "user", repair_prompt)
        await self._memory.append_history(session_id, "assistant", render.text)
        await self._enforce_history_limit(session_id)
        compaction_result = await self._compact_history_if_needed(
            session_id,
            prompt_cache_key=f"{channel}:{chat_id}:format-repair",
            system_prompt=system_prompt,
            notify=False,
        )
        turn_total_tokens += compaction_result.tokens_used
        metadata = self._response_metadata(True)
        metadata["format_repair_attempt"] = attempt
        metadata["format_repair_original_kind"] = original_kind
        metadata["token_trace"] = self._build_token_trace(
            turn_total_tokens=turn_total_tokens,
            session_total_tokens_before_compaction=compaction_result.session_total_tokens_before_compaction,
            session_total_tokens_after_compaction=compaction_result.session_total_tokens_after_compaction,
            compaction_performed=compaction_result.performed,
        )
        return ChannelResponse(
            channel=channel,
            chat_id=chat_id,
            text=render.text,
            render=render,
            metadata=metadata,
        )

    @staticmethod
    def _build_format_repair_prompt(
        *, channel: str, original_kind: str, parse_error: str, original_content: str
    ) -> str:
        return PromptService.build_format_repair_prompt(
            channel=channel,
            original_kind=original_kind,
            parse_error=parse_error,
            original_content=original_content,
        )

    async def _run_with_agent_runtime(
        self,
        *,
        session_id: str,
        history: list[Any],
        model_text: str,
        model_user_content: str | list[dict[str, Any]] | None,
        system_prompt: str,
        tool_context: ToolContext,
        prompt_cache_key: str | None,
        previous_response_id: str | None,
        chat_id: int | None,
        channel: str | None,
    ) -> _AgentRuntimeResult:
        if self._runtime is None:
            raise RuntimeError("agent runtime is not available")
        runtime_service_runtime = getattr(self._runtime_service, "_runtime", None)
        if self._runtime_service is None or runtime_service_runtime is not self._runtime:
            self._runtime_service = RuntimeOrchestrationService(
                runtime=self._runtime,
                llm_client=self._llm_client,
                guardrail=self._tool_use_guardrail,
                session_state=self._session_state,
                logger=self._logger,
            )
        assert self._runtime_service is not None
        result = await self._runtime_service.run_with_agent_runtime(
            session_id=session_id,
            history=history,
            model_text=model_text,
            model_user_content=model_user_content,
            system_prompt=system_prompt,
            tool_context=tool_context,
            prompt_cache_key=prompt_cache_key,
            previous_response_id=previous_response_id,
            chat_id=chat_id,
            channel=channel,
            response_schema=self._response_schema(),
        )
        return _AgentRuntimeResult(
            render=result.render,
            should_reply=result.should_reply,
            response_id=result.response_id,
            agent_trace=result.agent_trace,
            delegation_fallback_used=result.delegation_fallback_used,
            tokens_used=result.tokens_used,
        )

    def _build_model_user_input(self, message: ChannelMessage) -> tuple[str, str | list[dict[str, Any]] | None]:
        return self._input_service.build_model_user_input(message)

    def _response_schema(self) -> dict[str, Any]:
        return assistant_response_schema(kinds=["text", "html", "markdown_v2"], include_meta=True)

    async def _enforce_history_limit(self, session_id: str) -> None:
        if self._max_history_messages is None:
            return
        await self._memory.trim_history(session_id, self._max_history_messages)

    def _track_token_usage(self, session_id: str, tokens: int | None) -> int:
        return self._session_state.track_tokens(session_id, tokens)

    def _current_session_tokens(self, session_id: str) -> int:
        return self._session_state.current_tokens(session_id)

    @staticmethod
    def _build_token_trace(
        *,
        turn_total_tokens: int,
        session_total_tokens_before_compaction: int | None,
        session_total_tokens_after_compaction: int,
        compaction_performed: bool,
    ) -> dict[str, Any]:
        return SessionStateService.build_token_trace(
            turn_total_tokens=turn_total_tokens,
            session_total_tokens_before_compaction=session_total_tokens_before_compaction,
            session_total_tokens_after_compaction=session_total_tokens_after_compaction,
            compaction_performed=compaction_performed,
        )

    def _compact_system_prompt(self, system_prompt: str) -> str:
        return self._prompt_service.compact_system_prompt(system_prompt)

    async def _compact_history_if_needed(
        self,
        session_id: str,
        *,
        prompt_cache_key: str,
        system_prompt: str,
        notify: bool,
    ) -> _CompactionResult:
        result = await self._compaction_service.compact_history_if_needed(
            session_id,
            prompt_cache_key=prompt_cache_key,
            system_prompt=system_prompt,
            notify=notify,
            responses_state_mode=self._responses_state_mode(),
        )
        return _CompactionResult(
            updates=result.updates,
            performed=result.performed,
            tokens_used=result.tokens_used,
            session_total_tokens_before_compaction=result.session_total_tokens_before_compaction,
            session_total_tokens_after_compaction=result.session_total_tokens_after_compaction,
        )

    def _format_runtime_error_message(self, exc: Exception) -> str:
        if not self._logger.isEnabledFor(logging.DEBUG):
            return "Sorry, I couldn't answer right now."
        error_name = type(exc).__name__
        detail = str(exc).strip().replace("\n", " ")
        if detail:
            if len(detail) > 200:
                detail = f"{detail[:200]}..."
            return f"LLM error ({error_name}): {detail}"
        return f"LLM error ({error_name})."


def resolve_owner_id(message: ChannelMessage, default_owner_id: str | None) -> str:
    if default_owner_id:
        return default_owner_id
    if message.user_id is not None:
        return str(message.user_id)
    if message.chat_id is not None:
        return str(message.chat_id)
    return session_id_for(message)


def _prompt_cache_key(message: ChannelMessage) -> str | None:
    if message.channel and (message.user_id is not None or message.chat_id is not None):
        suffix = message.user_id if message.user_id is not None else message.chat_id
        return f"{message.channel}:{suffix}"
    session_key = session_id_for(message)
    if message.channel:
        return f"{message.channel}:{session_key}"
    return session_key



from __future__ import annotations

from typing import Any


async def complete_with_schema_fallback(
    *,
    provider: Any,
    call_kwargs: dict[str, Any],
    provider_name: str,
    model: str,
    provider_display_name: str,
    logger: Any,
) -> Any:
    try:
        return await provider.acomplete(**call_kwargs)
    except Exception as exc:
        if not should_retry_without_response_schema(call_kwargs=call_kwargs, exc=exc, provider_name=provider_name):
            raise
        retry_kwargs = dict(call_kwargs)
        retry_kwargs["response_schema"] = None
        logger.warning(
            "retrying openrouter request without response_schema",
            extra={"model": model, "provider": provider_display_name},
        )
        return await provider.acomplete(**retry_kwargs)


def should_retry_without_response_schema(*, call_kwargs: dict[str, Any], exc: Exception, provider_name: str) -> bool:
    if provider_name != "openrouter":
        return False
    if call_kwargs.get("response_schema") is None:
        return False
    message = str(exc).lower()
    if "json mode is not supported" in message:
        return True
    if '"code":20024' in message:
        return True
    return False



from __future__ import annotations

from dataclasses import dataclass
import logging
from typing import Sequence

from minibot.app.response_parser import extract_answer
from minibot.core.memory import MemoryBackend
from minibot.llm.provider_factory import LLMClient

from minibot.app.handlers.services.prompt_service import PromptService
from minibot.app.handlers.services.session_state_service import SessionStateService


@dataclass(frozen=True)
class CompactionResult:
    updates: list[str]
    performed: bool
    tokens_used: int
    session_total_tokens_before_compaction: int | None
    session_total_tokens_after_compaction: int


class HistoryCompactionService:
    def __init__(
        self,
        *,
        memory: MemoryBackend,
        llm_client: LLMClient,
        session_state: SessionStateService,
        prompt_service: PromptService,
        logger: logging.Logger,
        max_history_tokens: int | None,
        compaction_user_request: str,
    ) -> None:
        self._memory = memory
        self._llm_client = llm_client
        self._session_state = session_state
        self._prompt_service = prompt_service
        self._logger = logger
        self._max_history_tokens = max_history_tokens
        self._compaction_user_request = compaction_user_request

    async def compact_history_if_needed(
        self,
        session_id: str,
        *,
        prompt_cache_key: str,
        system_prompt: str,
        notify: bool,
        responses_state_mode: str,
    ) -> CompactionResult:
        updates: list[str] = []
        if self._max_history_tokens is None:
            return CompactionResult(
                updates=updates,
                performed=False,
                tokens_used=0,
                session_total_tokens_before_compaction=None,
                session_total_tokens_after_compaction=self._session_state.current_tokens(session_id),
            )
        pressure_tokens = self._session_state.current_tokens(session_id)
        if self._llm_client.is_responses_provider():
            latest_input_tokens = self._session_state.latest_input_tokens(session_id)
            if latest_input_tokens is not None:
                pressure_tokens = latest_input_tokens
        if pressure_tokens < self._max_history_tokens:
            return CompactionResult(
                updates=updates,
                performed=False,
                tokens_used=0,
                session_total_tokens_before_compaction=None,
                session_total_tokens_after_compaction=pressure_tokens,
            )
        history = list(await self._memory.get_history(session_id))
        if not history:
            session_before_reset = pressure_tokens
            self._session_state.session_total_tokens[session_id] = 0
            return CompactionResult(
                updates=updates,
                performed=False,
                tokens_used=0,
                session_total_tokens_before_compaction=session_before_reset,
                session_total_tokens_after_compaction=0,
            )
        if notify:
            updates.append("running compaction...")
        compaction_tokens = 0
        try:
            if self._should_use_responses_compaction_endpoint(session_id, responses_state_mode):
                try:
                    previous_response_id = self._session_state.get_previous_response_id(session_id)
                    if previous_response_id:
                        compacted = await self._llm_client.compact_response(
                            previous_response_id=previous_response_id,
                            prompt_cache_key=f"{prompt_cache_key}:compact",
                        )
                        compaction_text = self._extract_compaction_text(compacted.output) or "Conversation compacted."
                        session_before_reset = self._session_state.current_tokens(session_id)
                        compaction_tokens = self._session_state.track_tokens(session_id, compacted.total_tokens)
                        await self._memory.trim_history(session_id, 0)
                        await self._memory.append_history(session_id, "user", self._compaction_user_request)
                        await self._memory.append_history(session_id, "assistant", compaction_text)
                        self._session_state.session_total_tokens[session_id] = 0
                        self._session_state.set_previous_response_id(session_id, compacted.response_id)
                        if notify:
                            updates.append("done compacting")
                            updates.append(compaction_text)
                        return CompactionResult(
                            updates=updates,
                            performed=True,
                            tokens_used=compaction_tokens,
                            session_total_tokens_before_compaction=session_before_reset,
                            session_total_tokens_after_compaction=0,
                        )
                except Exception as exc:
                    self._logger.warning(
                        "responses compact endpoint failed; falling back to summary compaction",
                        extra={"error": str(exc)},
                    )

            compact_generation = await self._llm_client.generate(
                history,
                self._compaction_user_request,
                user_content=None,
                tools=[],
                tool_context=None,
                response_schema=None,
                prompt_cache_key=f"{prompt_cache_key}:compact",
                previous_response_id=None,
                system_prompt_override=self._prompt_service.compact_system_prompt(system_prompt),
            )
            session_before_reset = self._session_state.current_tokens(session_id)
            compaction_tokens = self._session_state.track_tokens(
                session_id,
                getattr(compact_generation, "total_tokens", None),
            )
            compact_render, _ = extract_answer(compact_generation.payload, logger=self._logger)
            await self._memory.trim_history(session_id, 0)
            await self._memory.append_history(session_id, "user", self._compaction_user_request)
            await self._memory.append_history(session_id, "assistant", compact_render.text)
            self._session_state.session_total_tokens[session_id] = 0
            if responses_state_mode == "previous_response_id":
                fallback_response_id = getattr(compact_generation, "response_id", None)
                if isinstance(fallback_response_id, str) and fallback_response_id:
                    self._session_state.set_previous_response_id(session_id, fallback_response_id)
                else:
                    self._session_state.clear_previous_response_id(session_id)
            if notify:
                updates.append("done compacting")
                updates.append(compact_render.text)
            return CompactionResult(
                updates=updates,
                performed=True,
                tokens_used=compaction_tokens,
                session_total_tokens_before_compaction=session_before_reset,
                session_total_tokens_after_compaction=0,
            )
        except Exception as exc:
            self._logger.exception("history compaction failed", exc_info=exc)
            if notify:
                updates.append("error compacting")
            return CompactionResult(
                updates=updates,
                performed=False,
                tokens_used=compaction_tokens,
                session_total_tokens_before_compaction=None,
                session_total_tokens_after_compaction=self._session_state.current_tokens(session_id),
            )

    def _should_use_responses_compaction_endpoint(self, session_id: str, responses_state_mode: str) -> bool:
        if not self._llm_client.is_responses_provider():
            return False
        if responses_state_mode != "previous_response_id":
            return False
        return self._session_state.has_previous_response_id(session_id)

    @staticmethod
    def _extract_compaction_text(items: Sequence[dict[str, object]]) -> str:
        parts: list[str] = []
        for item in items:
            if not isinstance(item, dict):
                continue
            content = item.get("content")
            if not isinstance(content, list):
                continue
            for entry in content:
                if not isinstance(entry, dict):
                    continue
                text = entry.get("text")
                if isinstance(text, str) and text:
                    parts.append(text)
        return "".join(parts).strip()



from minibot.app.handlers.services.compaction_service import CompactionResult, HistoryCompactionService
from minibot.app.handlers.services.input_service import UserInputService
from minibot.app.handlers.services.metadata_service import ResponseMetadataService
from minibot.app.handlers.services.prompt_service import PromptService
from minibot.app.handlers.services.runtime_service import AgentRuntimeResult, RuntimeOrchestrationService
from minibot.app.handlers.services.session_state_service import SessionStateService

__all__ = [
    "AgentRuntimeResult",
    "CompactionResult",
    "HistoryCompactionService",
    "PromptService",
    "ResponseMetadataService",
    "RuntimeOrchestrationService",
    "SessionStateService",
    "UserInputService",
]



from __future__ import annotations

from typing import Any, Sequence

from llm_async.models.tool_call import ToolCall

from minibot.llm.services.tool_executor import stringify_result, tool_name_from_call


MAX_REPEATED_TOOL_ITERATIONS = 3


def assistant_message_for_followup(message: Any) -> dict[str, Any]:
    payload: dict[str, Any] = {
        "role": getattr(message, "role", "assistant") or "assistant",
        "content": getattr(message, "content", "") or "",
    }
    tool_calls = getattr(message, "tool_calls", None)
    if tool_calls:
        payload["tool_calls"] = [tool_call_to_payload(call) for call in tool_calls]
    return payload


def tool_call_to_payload(call: ToolCall) -> dict[str, Any]:
    payload: dict[str, Any] = {
        "id": call.id,
        "type": call.type,
    }
    if call.function is not None:
        payload["function"] = call.function
    if call.name is not None:
        payload["name"] = call.name
    if call.input is not None:
        payload["input"] = call.input
    return payload


def tool_loop_fallback_payload(
    tool_messages: Sequence[dict[str, Any]],
    tool_names: Sequence[str],
    response_schema: dict[str, Any] | None,
) -> Any:
    summary = summarize_tool_outputs(tool_messages)
    tools_used = ", ".join(tool_names[-4:]) if tool_names else "tools"
    answer = (
        "I executed tool calls but hit an internal tool-loop safeguard before finalizing. "
        f"Recent tools: {tools_used}. Last tool output: {summary}"
    )
    if response_schema:
        return {
            "answer": answer,
            "should_answer_to_user": True,
        }
    return answer


def summarize_tool_outputs(tool_messages: Sequence[dict[str, Any]]) -> str:
    if not tool_messages:
        return "no tool output available"
    last = tool_messages[-1]
    output = last.get("output") if isinstance(last, dict) else None
    if output is None and isinstance(last, dict):
        output = last.get("content")
    if isinstance(output, str):
        return output[:400]
    return stringify_result(output)[:400]


def tool_iteration_signature(
    tool_calls: Sequence[ToolCall],
    tool_messages: Sequence[dict[str, Any]],
) -> str:
    parts: list[str] = []
    for index, call in enumerate(tool_calls):
        name = tool_name_from_call(call)
        output = ""
        if index < len(tool_messages):
            message = tool_messages[index]
            if isinstance(message, dict):
                output_value = message.get("output")
                if output_value is None:
                    output_value = message.get("content")
                if isinstance(output_value, str):
                    output = output_value[:240]
                else:
                    output = stringify_result(output_value)[:240]
        parts.append(f"{name}:{output}")
    return "|".join(parts)



from __future__ import annotations

from typing import Any, Awaitable, Callable

from minibot.llm.services.models import LLMCompaction, LLMGeneration
from minibot.llm.services.request_builder import RequestContext, build_continue_call_kwargs
from minibot.llm.services.usage_parser import (
    extract_response_id,
    extract_total_tokens_from_payload,
    extract_usage_from_response,
)
from minibot.shared.retries import AsyncRetriesService, RetryPolicy


async def compact_response(
    *,
    provider: Any,
    is_responses_provider: bool,
    model: str,
    prompt_cache_enabled: bool,
    previous_response_id: str,
    prompt_cache_key: str | None,
    retries_service: AsyncRetriesService,
    retry_attempts: int,
    retry_base_delay_seconds: float,
    retry_max_delay_seconds: float,
    logger: Any,
) -> LLMCompaction:
    if not is_responses_provider:
        raise RuntimeError("compaction endpoint is available only for openai_responses provider")
    payload: dict[str, Any] = {
        "model": model,
        "previous_response_id": previous_response_id,
    }
    if prompt_cache_key and prompt_cache_enabled:
        payload["prompt_cache_key"] = prompt_cache_key

    retry_policy = RetryPolicy(
        max_attempts=retry_attempts,
        base_delay_seconds=retry_base_delay_seconds,
        max_delay_seconds=retry_max_delay_seconds,
        backoff_factor=2.0,
        jitter=False,
        retry_exceptions=(Exception,),
    )

    def on_retry(exc: Exception, attempt: int, delay: float) -> None:
        logger.warning(
            "responses compact request failed; retrying",
            extra={
                "attempt": attempt,
                "next_delay_seconds": round(delay, 3),
                "model": model,
                "error": str(exc),
            },
        )

    raw = await retries_service.run(
        lambda: provider.request("POST", "/responses/compact", json_data=payload),
        policy=retry_policy,
        on_retry=on_retry,
    )
    response_id = raw.get("id") if isinstance(raw.get("id"), str) else ""
    if not response_id:
        raise RuntimeError("responses compact endpoint returned payload without id")
    output = raw.get("output")
    normalized_output = output if isinstance(output, list) else []
    total_tokens = extract_total_tokens_from_payload(raw)
    return LLMCompaction(
        response_id=response_id,
        output=normalized_output,
        total_tokens=total_tokens,
    )


async def continue_incomplete_response(
    *,
    complete_with_schema_fallback: Callable[[dict[str, Any]], Awaitable[Any]],
    ctx: RequestContext,
    previous_response_id: str,
    prompt_cache_key: str | None,
    system_prompt: str,
    response_schema: dict[str, Any] | None,
    logger: Any,
) -> LLMGeneration:
    call_kwargs = build_continue_call_kwargs(
        ctx=ctx,
        previous_response_id=previous_response_id,
        prompt_cache_key=prompt_cache_key,
        system_prompt=system_prompt,
        response_schema=response_schema,
    )
    logger.warning(
        "responses output incomplete; attempting one continuation",
        extra={"model": ctx.model, "response_id": previous_response_id},
    )
    response = await complete_with_schema_fallback(call_kwargs)
    message = response.main_response
    if not message:
        raise RuntimeError("LLM did not return a continuation completion")
    usage = extract_usage_from_response(response)
    return LLMGeneration(
        payload=message.content,
        response_id=extract_response_id(response),
        total_tokens=usage.total_tokens,
        input_tokens=usage.input_tokens,
        output_tokens=usage.output_tokens,
        cached_input_tokens=usage.cached_input_tokens,
        reasoning_output_tokens=usage.reasoning_output_tokens,
        status=usage.status,
        incomplete_reason=usage.incomplete_reason,
    )



from __future__ import annotations

import asyncio
from dataclasses import dataclass
from random import uniform
from typing import Awaitable, Callable, TypeVar


T = TypeVar("T")


@dataclass(frozen=True)
class RetryPolicy:
    max_attempts: int
    base_delay_seconds: float
    max_delay_seconds: float
    backoff_factor: float = 2.0
    jitter: bool = False
    retry_exceptions: tuple[type[Exception], ...] = (Exception,)


class AsyncRetriesService:
    async def run(
        self,
        operation: Callable[[], Awaitable[T]],
        *,
        policy: RetryPolicy,
        should_retry: Callable[[Exception], bool] | None = None,
        on_retry: Callable[[Exception, int, float], None] | None = None,
    ) -> T:
        if policy.max_attempts < 1:
            raise ValueError("max_attempts must be >= 1")
        if policy.base_delay_seconds <= 0:
            raise ValueError("base_delay_seconds must be > 0")
        if policy.max_delay_seconds <= 0:
            raise ValueError("max_delay_seconds must be > 0")
        if policy.backoff_factor < 1.0:
            raise ValueError("backoff_factor must be >= 1.0")

        last_error: Exception | None = None
        for attempt in range(1, policy.max_attempts + 1):
            try:
                return await operation()
            except Exception as exc:
                if not isinstance(exc, policy.retry_exceptions):
                    raise
                if should_retry is not None and not should_retry(exc):
                    raise
                if attempt >= policy.max_attempts:
                    raise
                last_error = exc
                delay = self._delay_for_attempt(policy=policy, attempt=attempt)
                if on_retry is not None:
                    on_retry(exc, attempt, delay)
                await asyncio.sleep(delay)

        assert last_error is not None
        raise last_error

    @staticmethod
    def _delay_for_attempt(*, policy: RetryPolicy, attempt: int) -> float:
        delay = min(policy.max_delay_seconds, policy.base_delay_seconds * (policy.backoff_factor ** (attempt - 1)))
        if policy.jitter:
            return uniform(delay * 0.5, delay)
        return delay



from __future__ import annotations

import logging
from typing import Any, Sequence

from llm_async.models.tool_call import ToolCall

from minibot.adapters.config.schema import LLMMConfig
from minibot.core.memory import MemoryEntry
from minibot.llm.services.client_bootstrap import (
    build_openrouter_provider_payload,
    create_provider,
    load_system_prompt,
    resolve_openrouter_reasoning_enabled,
)
from minibot.llm.services.compaction import compact_response as compact_response_via_service
from minibot.llm.services.compaction import continue_incomplete_response
from minibot.llm.services.models import LLMCompaction, LLMCompletionStep, LLMGeneration, ToolExecutionRecord
from minibot.llm.services.provider_registry import is_responses_provider_instance
from minibot.llm.services.request_builder import (
    RequestContext,
    build_complete_once_call_kwargs,
    build_generate_extra_kwargs,
    build_generate_step_call_kwargs,
    build_messages,
)
from minibot.llm.services.schema_fallback import complete_with_schema_fallback
from minibot.llm.services.schema_policy import normalize_response_schema, prepare_tool_specs
from minibot.llm.services.tool_executor import execute_tool_calls, execute_tool_calls_for_runtime, tool_name_from_call
from minibot.llm.services.tool_loop_guard import (
    MAX_REPEATED_TOOL_ITERATIONS,
    assistant_message_for_followup,
    tool_iteration_signature,
    tool_loop_fallback_payload,
)
from minibot.llm.services.usage_parser import (
    UsageAccumulator,
    extract_response_id,
    extract_total_tokens,
    extract_usage_from_response,
    parse_structured_payload,
    should_auto_continue_incomplete,
)
from minibot.llm.tools.base import ToolBinding, ToolContext
from minibot.shared.retries import AsyncRetriesService
from minibot.shared.utils import humanize_token_count


class LLMClient:
    def __init__(self, config: LLMMConfig) -> None:
        self._provider, self._provider_name = create_provider(config)
        self._model = config.model
        self._temperature = config.temperature
        self._max_new_tokens = config.max_new_tokens
        self._max_tool_iterations = config.max_tool_iterations
        self._system_prompt = load_system_prompt(config)
        self._prompts_dir = getattr(config, "prompts_dir", "./prompts")
        self._reasoning_effort = getattr(config, "reasoning_effort", "medium")
        self._responses_state_mode = getattr(config, "responses_state_mode", "full_messages")
        self._prompt_cache_enabled = bool(getattr(config, "prompt_cache_enabled", True))
        self._prompt_cache_retention = getattr(config, "prompt_cache_retention", None)
        self._compaction_retry_attempts = 3
        self._compaction_retry_base_delay_seconds = float(config.retry_delay_seconds)
        self._compaction_retry_max_delay_seconds = min(self._compaction_retry_base_delay_seconds * 4, 10.0)
        self._retries_service = AsyncRetriesService()
        self._openrouter_models = tuple(getattr(getattr(config, "openrouter", None), "models", []) or [])
        self._openrouter_provider = build_openrouter_provider_payload(config)
        self._openrouter_reasoning_enabled = resolve_openrouter_reasoning_enabled(config)
        self._openrouter_plugins = tuple(getattr(getattr(config, "openrouter", None), "plugins", []) or [])
        self._is_responses_provider = is_responses_provider_instance(self._provider)
        self._logger = logging.getLogger("minibot.llm")

    async def generate(
        self,
        history: Sequence[MemoryEntry],
        user_message: str,
        user_content: str | list[dict[str, Any]] | None = None,
        tools: Sequence[ToolBinding] | None = None,
        tool_context: ToolContext | None = None,
        response_schema: dict[str, Any] | None = None,
        prompt_cache_key: str | None = None,
        previous_response_id: str | None = None,
        system_prompt_override: str | None = None,
    ) -> LLMGeneration:
        system_prompt = system_prompt_override or self._system_prompt
        messages = build_messages(
            history=history,
            user_message=user_message,
            user_content=user_content,
            system_prompt=system_prompt,
        )

        if not self._provider.api_key:
            self._logger.warning("LLM provider key missing, falling back to echo", extra={"component": "llm"})
            return LLMGeneration(f"Echo: {user_message}")

        conversation = list(messages)
        tool_bindings = list(tools or [])
        tool_specs = prepare_tool_specs(tool_bindings, self._model)
        context = tool_context or ToolContext()
        iterations = 0
        last_tool_messages: list[dict[str, Any]] = []
        recent_tool_names: list[str] = []
        last_iteration_signature: str | None = None
        repeated_iteration_count = 0
        strict_response_schema = normalize_response_schema(response_schema, self._model)
        request_ctx = self._request_context()
        extra_kwargs = build_generate_extra_kwargs(
            ctx=request_ctx,
            prompt_cache_key=prompt_cache_key,
            previous_response_id=previous_response_id,
            system_prompt=system_prompt,
        )
        usage_accumulator = UsageAccumulator()

        while True:
            call_kwargs = build_generate_step_call_kwargs(
                ctx=request_ctx,
                conversation=conversation,
                tool_specs=tool_specs,
                strict_response_schema=strict_response_schema,
                extra_kwargs=extra_kwargs,
            )
            response = await self._complete_with_schema_fallback(call_kwargs)
            usage = extract_usage_from_response(response)
            usage_tokens = extract_total_tokens(response)
            usage_accumulator.add_step(usage, usage_tokens)

            message = response.main_response
            if not message:
                raise RuntimeError("LLM did not return a completion")
            self._logger.debug(
                "llm completion received",
                extra={
                    "tool_calls": len(getattr(message, "tool_calls", None) or []),
                    "step_tokens": humanize_token_count(usage_tokens) if isinstance(usage_tokens, int) else "0",
                    "cumulative_tokens": humanize_token_count(usage_accumulator.total_tokens_used),
                    "provider": self.provider_name(),
                    "response_status": usage.status,
                    "incomplete_reason": usage.incomplete_reason,
                    "applied_max_output_tokens": self._max_new_tokens if self._is_responses_provider else None,
                },
            )
            if not message.tool_calls or not tool_bindings:
                payload = message.content
                response_id = extract_response_id(response)
                status = usage.status
                incomplete_reason = usage.incomplete_reason
                if self._is_responses_provider and response_id and should_auto_continue_incomplete(usage):
                    continuation = await continue_incomplete_response(
                        complete_with_schema_fallback=self._complete_with_schema_fallback,
                        ctx=request_ctx,
                        previous_response_id=response_id,
                        prompt_cache_key=prompt_cache_key,
                        system_prompt=system_prompt,
                        response_schema=strict_response_schema,
                        logger=self._logger,
                    )
                    continuation_payload = (
                        continuation.payload if isinstance(continuation.payload, str) else str(continuation.payload)
                    )
                    payload = f"{payload}{continuation_payload}"
                    usage_accumulator.add_generation(continuation)
                    response_id = continuation.response_id or response_id
                    status = continuation.status
                    incomplete_reason = continuation.incomplete_reason
                if response_schema and isinstance(payload, str):
                    try:
                        parsed = parse_structured_payload(payload)
                        return usage_accumulator.build_generation(
                            payload=parsed,
                            response_id=response_id,
                            status=status,
                            incomplete_reason=incomplete_reason,
                        )
                    except Exception:
                        self._logger.warning("failed to parse structured response; falling back to text")
                return usage_accumulator.build_generation(
                    payload=payload,
                    response_id=response_id,
                    status=status,
                    incomplete_reason=incomplete_reason,
                )

            tool_messages = await execute_tool_calls(
                message.tool_calls,
                tool_bindings,
                context,
                responses_mode=self._is_responses_provider,
                logger=self._logger,
            )
            iteration_signature = tool_iteration_signature(message.tool_calls, tool_messages)
            if iteration_signature and iteration_signature == last_iteration_signature:
                repeated_iteration_count += 1
            else:
                repeated_iteration_count = 1
            last_iteration_signature = iteration_signature
            last_tool_messages = tool_messages
            recent_tool_names.extend(tool_name_from_call(call) for call in message.tool_calls)
            if repeated_iteration_count >= MAX_REPEATED_TOOL_ITERATIONS:
                self._logger.warning(
                    "tool loop repeated identical outputs; returning fallback",
                    extra={
                        "tool_names": recent_tool_names[-10:],
                        "repeated_count": repeated_iteration_count,
                    },
                )
                response_id = extract_response_id(response)
                payload = tool_loop_fallback_payload(last_tool_messages, recent_tool_names, response_schema)
                return LLMGeneration(
                    payload,
                    response_id,
                    total_tokens=usage_accumulator.total_tokens_used or None,
                )
            if self._is_responses_provider:
                response_id = extract_response_id(response)
                if response_id:
                    extra_kwargs["previous_response_id"] = response_id
                conversation = tool_messages
            else:
                conversation.append(assistant_message_for_followup(message))
                conversation.extend(tool_messages)
            iterations += 1
            if iterations >= self._max_tool_iterations:
                self._logger.warning(
                    "tool call loop exceeded maximum iterations; returning fallback",
                    extra={"tool_names": recent_tool_names[-10:]},
                )
                response_id = extract_response_id(response)
                payload = tool_loop_fallback_payload(last_tool_messages, recent_tool_names, response_schema)
                return LLMGeneration(
                    payload,
                    response_id,
                    total_tokens=usage_accumulator.total_tokens_used or None,
                )

    def is_responses_provider(self) -> bool:
        return self._is_responses_provider

    def responses_state_mode(self) -> str:
        if self._responses_state_mode in {"full_messages", "previous_response_id"}:
            return self._responses_state_mode
        return "full_messages"

    def prompt_cache_enabled(self) -> bool:
        return self._prompt_cache_enabled

    async def compact_response(
        self,
        *,
        previous_response_id: str,
        prompt_cache_key: str | None = None,
    ) -> LLMCompaction:
        return await compact_response_via_service(
            provider=self._provider,
            is_responses_provider=self._is_responses_provider,
            model=self._model,
            prompt_cache_enabled=self._prompt_cache_enabled,
            previous_response_id=previous_response_id,
            prompt_cache_key=prompt_cache_key,
            retries_service=self._retries_service,
            retry_attempts=self._compaction_retry_attempts,
            retry_base_delay_seconds=self._compaction_retry_base_delay_seconds,
            retry_max_delay_seconds=self._compaction_retry_max_delay_seconds,
            logger=self._logger,
        )

    async def complete_once(
        self,
        messages: Sequence[dict[str, Any]],
        tools: Sequence[ToolBinding] | None = None,
        response_schema: dict[str, Any] | None = None,
        prompt_cache_key: str | None = None,
        previous_response_id: str | None = None,
    ) -> LLMCompletionStep:
        strict_response_schema = normalize_response_schema(response_schema, self._model)
        tool_specs = prepare_tool_specs(tools or [], self._model)
        request_ctx = self._request_context()
        call_kwargs = build_complete_once_call_kwargs(
            ctx=request_ctx,
            messages=messages,
            tool_specs=tool_specs,
            strict_response_schema=strict_response_schema,
            prompt_cache_key=prompt_cache_key,
            previous_response_id=previous_response_id,
        )

        response = await self._complete_with_schema_fallback(call_kwargs)
        message = response.main_response
        if not message:
            raise RuntimeError("LLM did not return a completion")
        usage_tokens = extract_total_tokens(response)
        self._logger.debug(
            "llm runtime completion step received",
            extra={
                "tool_calls": len(getattr(message, "tool_calls", None) or []),
                "step_tokens": humanize_token_count(usage_tokens) if isinstance(usage_tokens, int) else "0",
                "provider": self.provider_name(),
            },
        )
        return LLMCompletionStep(
            message=message,
            response_id=extract_response_id(response),
            total_tokens=usage_tokens,
        )

    async def execute_tool_calls_for_runtime(
        self,
        tool_calls: Sequence[ToolCall],
        tools: Sequence[ToolBinding],
        context: ToolContext,
        responses_mode: bool = False,
    ) -> list[ToolExecutionRecord]:
        return await execute_tool_calls_for_runtime(
            tool_calls,
            tools,
            context,
            responses_mode=responses_mode,
            logger=self._logger,
        )

    def provider_name(self) -> str:
        return self._provider_name

    def model_name(self) -> str:
        return self._model

    def system_prompt(self) -> str:
        return self._system_prompt

    def prompts_dir(self) -> str:
        return self._prompts_dir

    def max_tool_iterations(self) -> int:
        return self._max_tool_iterations

    def supports_media_inputs(self) -> bool:
        return self._provider_name in {"openai_responses", "openai", "openrouter"}

    def media_input_mode(self) -> str:
        if self._provider_name == "openai_responses":
            return "responses"
        if self._provider_name in {"openai", "openrouter"}:
            return "chat_completions"
        return "none"

    async def _complete_with_schema_fallback(self, call_kwargs: dict[str, Any]) -> Any:
        return await complete_with_schema_fallback(
            provider=self._provider,
            call_kwargs=call_kwargs,
            provider_name=self._provider_name,
            model=self._model,
            provider_display_name=self.provider_name(),
            logger=self._logger,
        )

    def _request_context(self) -> RequestContext:
        return RequestContext(
            model=self._model,
            provider_name=self._provider_name,
            is_responses_provider=self._is_responses_provider,
            temperature=self._temperature,
            max_new_tokens=self._max_new_tokens,
            prompt_cache_enabled=self._prompt_cache_enabled,
            prompt_cache_retention=self._prompt_cache_retention,
            reasoning_effort=self._reasoning_effort,
            openrouter_models=self._openrouter_models,
            openrouter_provider=self._openrouter_provider,
            openrouter_reasoning_enabled=self._openrouter_reasoning_enabled,
            openrouter_plugins=self._openrouter_plugins,
        )



from __future__ import annotations

from typing import Any, Sequence

from minibot.core.channels import ChannelMessage, IncomingFileRef


def incoming_files_from_metadata(metadata: dict[str, Any] | None) -> list[IncomingFileRef]:
    if not isinstance(metadata, dict):
        return []
    raw = metadata.get("incoming_files")
    if not isinstance(raw, list):
        return []
    parsed: list[IncomingFileRef] = []
    for item in raw:
        if not isinstance(item, dict):
            continue
        try:
            parsed.append(IncomingFileRef.model_validate(item))
        except Exception:
            continue
    return parsed


def summarize_attachments_for_memory(attachments: Sequence[dict[str, Any]]) -> str:
    summaries: list[str] = []
    for attachment in attachments:
        attachment_type = attachment.get("type")
        if attachment_type == "input_image":
            summaries.append("image")
            continue
        if attachment_type == "input_file":
            filename = attachment.get("filename")
            if isinstance(filename, str) and filename.strip():
                summaries.append(f"file:{filename.strip()}")
            else:
                summaries.append("file")
            continue
        summaries.append("attachment")
    return ", ".join(summaries)


def summarize_incoming_files_for_memory(incoming_files: Sequence[IncomingFileRef]) -> str:
    if not incoming_files:
        return ""
    return ", ".join([f"file:{item.filename}" for item in incoming_files])


def build_history_user_entry(message: ChannelMessage, model_text: str) -> str:
    base_text = message.text.strip() if message.text else ""
    attachment_summary = summarize_attachments_for_memory(message.attachments)
    incoming_files = incoming_files_from_metadata(message.metadata)
    incoming_file_summary = summarize_incoming_files_for_memory(incoming_files)
    if not attachment_summary and not incoming_file_summary:
        return base_text
    visible_text = base_text or model_text
    parts = [item for item in [attachment_summary, incoming_file_summary] if item]
    summary = ", ".join(parts)
    if visible_text:
        return f"{visible_text}\nAttachments: {summary}"
    return f"Attachments: {summary}"


def suggest_persist_destination(path: str) -> str | None:
    marker = "uploads/temp/"
    if path.startswith(marker):
        return f"uploads/{path[len(marker) :]}"
    return None


def build_incoming_files_text(prompt_text: str, incoming_files: Sequence[IncomingFileRef]) -> str:
    lines = [
        "Incoming managed files:",
        *[
            (
                f"- {item.filename} (path={item.path}, mime={item.mime}, size={item.size_bytes} bytes"
                f", source={item.source}, caption={item.caption or ''})"
            )
            for item in incoming_files
        ],
    ]
    first_path = incoming_files[0].path if incoming_files else ""
    suggested_destination = suggest_persist_destination(first_path)
    lines.append(
        "For file-management requests, use the filesystem tool "
        "(action=move, action=delete, action=send, action=list). "
        "Do NOT call self_insert_artifact unless user explicitly asks to inspect content."
    )
    if suggested_destination:
        lines.append(
            "If user asks to save the uploaded file, use filesystem action=move "
            f"source_path={first_path} destination_path={suggested_destination}."
        )
    lines.append("For analysis requests, use self_insert_artifact when inspection is required.")
    lines.append("If user intent is unclear, ask a clarifying question before acting.")
    if prompt_text:
        return f"{prompt_text}\n\n" + "\n".join(lines)
    return (
        "The user uploaded file(s) but did not include a clear instruction.\n"
        + "\n".join(lines)
        + "\nAsk the user what to do, unless the intent is already obvious."
    )



from minibot.llm.services.models import LLMCompaction, LLMCompletionStep, LLMGeneration, ToolExecutionRecord

__all__ = [
    "LLMGeneration",
    "LLMCompletionStep",
    "ToolExecutionRecord",
    "LLMCompaction",
]



from __future__ import annotations

import logging
from typing import Sequence

from minibot.llm.provider_factory import LLMClient
from minibot.llm.tools.base import ToolBinding
from minibot.shared.prompt_loader import load_channel_prompt, load_compact_prompt, load_policy_prompts


class PromptService:
    def __init__(
        self,
        llm_client: LLMClient,
        tools: Sequence[ToolBinding],
        environment_prompt_fragment: str,
        logger: logging.Logger,
    ) -> None:
        self._llm_client = llm_client
        self._tools = list(tools)
        self._environment_prompt_fragment = environment_prompt_fragment.strip()
        self._logger = logger
        self._prompts_dir = self._resolve_prompts_dir()

    @property
    def prompts_dir(self) -> str:
        return self._prompts_dir

    def compose_system_prompt(self, channel: str | None) -> str:
        system_prompt_getter = getattr(self._llm_client, "system_prompt", None)
        base_prompt = "You are Minibot, a helpful assistant."
        if callable(system_prompt_getter):
            maybe_prompt = system_prompt_getter()
            if isinstance(maybe_prompt, str) and maybe_prompt:
                base_prompt = maybe_prompt

        fragments = [base_prompt]
        fragments.extend(load_policy_prompts(self._prompts_dir))
        channel_prompt = load_channel_prompt(self._prompts_dir, channel)
        if channel_prompt:
            fragments.append(channel_prompt)
        if self._environment_prompt_fragment:
            fragments.append(self._environment_prompt_fragment)
        self._logger.debug(
            "composed system prompt",
            extra={
                "channel": channel,
                "prompts_dir": self._prompts_dir,
                "channel_prompt_loaded": bool(channel_prompt),
                "fragment_count": len(fragments),
                "prompt_preview": "\n\n".join(fragments)[:200],
            },
        )

        if any(binding.tool.name == "self_insert_artifact" for binding in self._tools):
            fragments.append(
                "When you need to inspect a local workspace file (image/document), call self_insert_artifact first "
                "to inject it into conversation context before answering file contents. "
                "For file-management requests (save, move, delete, send, list), do not call self_insert_artifact; "
                "use the filesystem tool with the appropriate action instead. "
                "If the user only uploaded files and gave no clear instruction, ask a clarifying question."
            )
        return "\n\n".join(fragments)

    def compact_system_prompt(self, system_prompt: str) -> str:
        compact_prompt = load_compact_prompt(self._prompts_dir)
        if compact_prompt:
            return f"{system_prompt}\n\n{compact_prompt}"
        return (
            f"{system_prompt}\n\n"
            "You are compacting conversation memory. Return a concise but complete summary of the "
            "conversation so far, preserving user goals, constraints, and pending tasks. "
            "Do not include preamble."
        )

    @staticmethod
    def build_format_repair_prompt(
        *,
        channel: str,
        original_kind: str,
        parse_error: str,
        original_content: str,
    ) -> str:
        if channel == "telegram":
            return (
                "We tried to send a formatted response to Telegram but got a formatting parse error. "
                "Rewrite the same answer with valid Telegram-compatible formatting.\n\n"
                f"Original kind: {original_kind}\n"
                f"Telegram error: {parse_error}\n\n"
                "Requirements:\n"
                "- Return the same meaning and content, only fix formatting.\n"
                "- Keep kind as markdown_v2 or html only if valid for Telegram, otherwise use text.\n"
                "- For markdown_v2, write normal Markdown (do not pre-escape Telegram MarkdownV2).\n"
                "- Do not use placeholder statements.\n"
                "- Return structured output only.\n\n"
                f"Original content:\n{original_content}"
            )
        return (
            "We tried to send a formatted response to the target channel and got a formatting parse error. "
            "Rewrite the same answer with valid channel-compatible formatting.\n\n"
            f"Channel: {channel}\n"
            f"Original kind: {original_kind}\n"
            f"Parse error: {parse_error}\n\n"
            "Requirements:\n"
            "- Return the same meaning and content, only fix formatting.\n"
            "- Keep kind aligned with valid formatting for this channel, otherwise use text.\n"
            "- Do not use placeholder statements.\n"
            "- Return structured output only.\n\n"
            f"Original content:\n{original_content}"
        )

    def _resolve_prompts_dir(self) -> str:
        prompts_dir_getter = getattr(self._llm_client, "prompts_dir", None)
        if callable(prompts_dir_getter):
            value = prompts_dir_getter()
            if isinstance(value, str) and value:
                return value
        return "./prompts"



from __future__ import annotations

from typing import Any

from llm_async.providers import ClaudeProvider, GoogleProvider, OpenAIProvider, OpenRouterProvider
from llm_async.providers.openai_responses import OpenAIResponsesProvider


LLM_PROVIDERS = {
    "openai": OpenAIProvider,
    "claude": ClaudeProvider,
    "google": GoogleProvider,
    "openrouter": OpenRouterProvider,
    "openai_responses": OpenAIResponsesProvider,
}


def resolve_provider_class(configured_provider: str) -> tuple[type[Any], str]:
    provider_cls = LLM_PROVIDERS.get(configured_provider, OpenAIProvider)
    provider_name = configured_provider if configured_provider in LLM_PROVIDERS else "openai"
    return provider_cls, provider_name


def is_responses_provider_instance(provider: Any) -> bool:
    return isinstance(provider, OpenAIResponsesProvider)



from __future__ import annotations

from typing import Any, Sequence

from minibot.app.incoming_files_context import build_incoming_files_text, incoming_files_from_metadata
from minibot.core.channels import ChannelMessage
from minibot.llm.provider_factory import LLMClient


class UserInputService:
    def __init__(self, llm_client: LLMClient) -> None:
        self._llm_client = llm_client

    def supports_media_inputs(self) -> bool:
        supports_getter = getattr(self._llm_client, "supports_media_inputs", None)
        if callable(supports_getter):
            return bool(supports_getter())
        return self._llm_client.is_responses_provider()

    def media_input_mode(self) -> str:
        mode_getter = getattr(self._llm_client, "media_input_mode", None)
        if callable(mode_getter):
            mode = mode_getter()
            if isinstance(mode, str) and mode:
                return mode
        return "responses" if self._llm_client.is_responses_provider() else "none"

    def build_model_user_input(
        self,
        message: ChannelMessage,
    ) -> tuple[str, str | list[dict[str, Any]] | None]:
        prompt_text = message.text.strip() if message.text else ""
        incoming_files = incoming_files_from_metadata(message.metadata)
        if incoming_files and not message.attachments:
            return build_incoming_files_text(prompt_text, incoming_files), None
        if not message.attachments:
            return prompt_text, None

        resolved_prompt = prompt_text or "Please analyze the attached media and summarize the key information."
        mode = self.media_input_mode()
        parts: list[dict[str, Any]] = []
        if mode == "chat_completions":
            parts.append({"type": "text", "text": resolved_prompt})
        else:
            parts.append({"type": "input_text", "text": resolved_prompt})
        parts.extend(self._transform_attachments_for_mode(message.attachments, mode))
        return resolved_prompt, parts

    def _transform_attachments_for_mode(
        self,
        attachments: Sequence[dict[str, Any]],
        mode: str,
    ) -> list[dict[str, Any]]:
        if mode == "chat_completions":
            return [self._to_chat_completions_attachment(attachment) for attachment in attachments]
        return [dict(attachment) for attachment in attachments]

    @staticmethod
    def _to_chat_completions_attachment(attachment: dict[str, Any]) -> dict[str, Any]:
        attachment_type = attachment.get("type")
        if attachment_type == "input_image":
            image_url = attachment.get("image_url")
            return {
                "type": "image_url",
                "image_url": {
                    "url": image_url,
                },
            }
        if attachment_type == "input_file":
            return {
                "type": "file",
                "file": {
                    "filename": attachment.get("filename"),
                    "file_data": attachment.get("file_data"),
                },
            }
        return dict(attachment)



from __future__ import annotations

from typing import Any

from minibot.llm.provider_factory import LLMClient


class ResponseMetadataService:
    def __init__(self, llm_client: LLMClient) -> None:
        self._llm_client = llm_client

    def provider_name(self) -> str | None:
        provider_getter = getattr(self._llm_client, "provider_name", None)
        if callable(provider_getter):
            provider = provider_getter()
            if isinstance(provider, str) and provider:
                return provider
        return None

    def model_name(self) -> str | None:
        model_getter = getattr(self._llm_client, "model_name", None)
        if callable(model_getter):
            model = model_getter()
            if isinstance(model, str) and model:
                return model
        return None

    def response_metadata(self, should_reply: bool) -> dict[str, Any]:
        return {
            "should_reply": should_reply,
            "llm_provider": self.provider_name(),
            "llm_model": self.model_name(),
        }



from __future__ import annotations

from typing import Any


class SessionStateService:
    def __init__(self) -> None:
        self.session_total_tokens: dict[str, int] = {}
        self.session_previous_response_ids: dict[str, str] = {}
        self.session_latest_input_tokens: dict[str, int] = {}
        self.session_latest_output_tokens: dict[str, int] = {}
        self.session_latest_total_tokens: dict[str, int] = {}
        self.session_latest_cached_input_tokens: dict[str, int] = {}
        self.session_latest_reasoning_output_tokens: dict[str, int] = {}

    def track_tokens(self, session_id: str, tokens: int | None) -> int:
        if tokens is None or tokens <= 0:
            return 0
        self.session_total_tokens[session_id] = self.session_total_tokens.get(session_id, 0) + tokens
        return tokens

    def current_tokens(self, session_id: str) -> int:
        return self.session_total_tokens.get(session_id, 0)

    def set_previous_response_id(self, session_id: str, response_id: str | None) -> None:
        if response_id:
            self.session_previous_response_ids[session_id] = response_id

    def get_previous_response_id(self, session_id: str) -> str | None:
        return self.session_previous_response_ids.get(session_id)

    def clear_previous_response_id(self, session_id: str) -> None:
        self.session_previous_response_ids.pop(session_id, None)

    def has_previous_response_id(self, session_id: str) -> bool:
        return session_id in self.session_previous_response_ids

    def track_usage(
        self,
        session_id: str,
        *,
        input_tokens: int | None,
        output_tokens: int | None,
        total_tokens: int | None,
        cached_input_tokens: int | None,
        reasoning_output_tokens: int | None,
    ) -> None:
        if input_tokens is not None and input_tokens >= 0:
            self.session_latest_input_tokens[session_id] = input_tokens
        if output_tokens is not None and output_tokens >= 0:
            self.session_latest_output_tokens[session_id] = output_tokens
        if total_tokens is not None and total_tokens >= 0:
            self.session_latest_total_tokens[session_id] = total_tokens
        if cached_input_tokens is not None and cached_input_tokens >= 0:
            self.session_latest_cached_input_tokens[session_id] = cached_input_tokens
        if reasoning_output_tokens is not None and reasoning_output_tokens >= 0:
            self.session_latest_reasoning_output_tokens[session_id] = reasoning_output_tokens

    def latest_input_tokens(self, session_id: str) -> int | None:
        return self.session_latest_input_tokens.get(session_id)

    def latest_usage_trace(self, session_id: str) -> dict[str, int | None]:
        return {
            "input_tokens": self.session_latest_input_tokens.get(session_id),
            "output_tokens": self.session_latest_output_tokens.get(session_id),
            "total_tokens": self.session_latest_total_tokens.get(session_id),
            "cached_input_tokens": self.session_latest_cached_input_tokens.get(session_id),
            "reasoning_output_tokens": self.session_latest_reasoning_output_tokens.get(session_id),
        }

    @staticmethod
    def build_token_trace(
        *,
        turn_total_tokens: int,
        session_total_tokens_before_compaction: int | None,
        session_total_tokens_after_compaction: int,
        compaction_performed: bool,
    ) -> dict[str, Any]:
        return {
            "turn_total_tokens": max(0, int(turn_total_tokens)),
            "session_total_tokens": max(0, int(session_total_tokens_after_compaction)),
            "session_total_tokens_before_compaction": session_total_tokens_before_compaction,
            "session_total_tokens_after_compaction": max(0, int(session_total_tokens_after_compaction)),
            "compaction_performed": compaction_performed,
            "accounting_scope": "all_turn_calls",
        }



from __future__ import annotations

import re
from typing import Any, Sequence

from llm_async.models import Tool

from minibot.llm.tools.base import ToolBinding
from minibot.shared.json_schema import to_openai_strict_schema


_OPENAI_STRICT_MODEL_PATTERNS = (
    re.compile(r"^openai(?:/.*)?"),
    re.compile(r"^gpt-.*"),
)


def should_apply_openai_strict_schema(model_name: str | None) -> bool:
    if not isinstance(model_name, str) or not model_name:
        return False
    return any(pattern.match(model_name) for pattern in _OPENAI_STRICT_MODEL_PATTERNS)


def normalize_response_schema(
    response_schema: dict[str, Any] | None,
    model_name: str | None,
) -> dict[str, Any] | None:
    if isinstance(response_schema, dict) and should_apply_openai_strict_schema(model_name):
        return to_openai_strict_schema(response_schema)
    return response_schema


def prepare_tool_specs(tool_bindings: Sequence[ToolBinding], model_name: str | None) -> list[Tool] | None:
    if not tool_bindings:
        return None
    if not should_apply_openai_strict_schema(model_name):
        return [binding.tool for binding in tool_bindings]

    strict_tools: list[Tool] = []
    for binding in tool_bindings:
        parameters = binding.tool.parameters
        if isinstance(parameters, dict):
            parameters = to_openai_strict_schema(parameters)
        strict_tools.append(
            Tool(
                name=binding.tool.name,
                description=binding.tool.description,
                parameters=parameters,
            )
        )
    return strict_tools



from __future__ import annotations

from dataclasses import dataclass
from typing import Any

from minibot.llm.services.models import LLMGeneration, UsageSnapshot
from minibot.shared.parse_utils import parse_json_with_fenced_fallback


def parse_structured_payload(payload: str) -> Any:
    return parse_json_with_fenced_fallback(payload)


def extract_response_id(response: Any) -> str | None:
    original = getattr(response, "original", None)
    if isinstance(original, dict):
        resp_id = original.get("id")
        if isinstance(resp_id, str):
            return resp_id
    return None


def extract_total_tokens(response: Any) -> int | None:
    original = getattr(response, "original", None)
    if not isinstance(original, dict):
        return None
    return extract_total_tokens_from_payload(original)


def extract_usage_from_response(response: Any) -> UsageSnapshot:
    original = getattr(response, "original", None)
    if not isinstance(original, dict):
        return UsageSnapshot()
    return extract_usage_from_payload(original)


def extract_total_tokens_from_payload(original: dict[str, Any]) -> int | None:
    return extract_usage_from_payload(original).total_tokens


def extract_usage_from_payload(original: dict[str, Any]) -> UsageSnapshot:
    usage = original.get("usage")
    if not isinstance(usage, dict):
        return UsageSnapshot(
            status=opt_str(original.get("status")),
            incomplete_reason=incomplete_reason(original),
        )
    total_tokens = opt_int(usage.get("total_tokens"))
    input_tokens = opt_int(usage.get("input_tokens"))
    output_tokens = opt_int(usage.get("output_tokens"))
    if total_tokens is None and input_tokens is not None and output_tokens is not None:
        total_tokens = input_tokens + output_tokens
    if total_tokens is None:
        prompt_tokens = opt_int(usage.get("prompt_tokens"))
        completion_tokens = opt_int(usage.get("completion_tokens"))
        if prompt_tokens is not None and completion_tokens is not None:
            total_tokens = prompt_tokens + completion_tokens
        if input_tokens is None:
            input_tokens = prompt_tokens
        if output_tokens is None:
            output_tokens = completion_tokens

    cached_input_tokens = None
    input_details = usage.get("input_tokens_details")
    if isinstance(input_details, dict):
        cached_input_tokens = opt_int(input_details.get("cached_tokens"))

    reasoning_output_tokens = None
    output_details = usage.get("output_tokens_details")
    if isinstance(output_details, dict):
        reasoning_output_tokens = opt_int(output_details.get("reasoning_tokens"))

    return UsageSnapshot(
        total_tokens=total_tokens,
        input_tokens=input_tokens,
        output_tokens=output_tokens,
        cached_input_tokens=cached_input_tokens,
        reasoning_output_tokens=reasoning_output_tokens,
        status=opt_str(original.get("status")),
        incomplete_reason=incomplete_reason(original),
    )


def should_auto_continue_incomplete(usage: UsageSnapshot) -> bool:
    if usage.status != "incomplete":
        return False
    reason = (usage.incomplete_reason or "").strip().lower()
    if not reason:
        return False
    return reason in {"max_output_tokens", "max_tokens"} or "max_output" in reason


@dataclass
class UsageAccumulator:
    total_tokens_used: int = 0
    input_tokens_used: int = 0
    output_tokens_used: int = 0
    cached_input_tokens_used: int = 0
    reasoning_output_tokens_used: int = 0
    saw_input_tokens: bool = False
    saw_output_tokens: bool = False
    saw_cached_input_tokens: bool = False
    saw_reasoning_output_tokens: bool = False

    def add_step(self, usage: UsageSnapshot, usage_tokens: int | None) -> None:
        if usage_tokens is not None:
            self.total_tokens_used += usage_tokens
        if usage.input_tokens is not None:
            self.input_tokens_used += usage.input_tokens
            self.saw_input_tokens = True
        if usage.output_tokens is not None:
            self.output_tokens_used += usage.output_tokens
            self.saw_output_tokens = True
        if usage.cached_input_tokens is not None:
            self.cached_input_tokens_used += usage.cached_input_tokens
            self.saw_cached_input_tokens = True
        if usage.reasoning_output_tokens is not None:
            self.reasoning_output_tokens_used += usage.reasoning_output_tokens
            self.saw_reasoning_output_tokens = True

    def add_generation(self, generation: LLMGeneration) -> None:
        if generation.total_tokens is not None:
            self.total_tokens_used += generation.total_tokens
        if generation.input_tokens is not None:
            self.input_tokens_used += generation.input_tokens
            self.saw_input_tokens = True
        if generation.output_tokens is not None:
            self.output_tokens_used += generation.output_tokens
            self.saw_output_tokens = True
        if generation.cached_input_tokens is not None:
            self.cached_input_tokens_used += generation.cached_input_tokens
            self.saw_cached_input_tokens = True
        if generation.reasoning_output_tokens is not None:
            self.reasoning_output_tokens_used += generation.reasoning_output_tokens
            self.saw_reasoning_output_tokens = True

    def build_generation(
        self,
        *,
        payload: Any,
        response_id: str | None,
        status: str | None,
        incomplete_reason: str | None,
    ) -> LLMGeneration:
        return LLMGeneration(
            payload,
            response_id,
            total_tokens=self.total_tokens_used or None,
            input_tokens=self.input_tokens_used if self.saw_input_tokens else None,
            output_tokens=self.output_tokens_used if self.saw_output_tokens else None,
            cached_input_tokens=self.cached_input_tokens_used if self.saw_cached_input_tokens else None,
            reasoning_output_tokens=self.reasoning_output_tokens_used if self.saw_reasoning_output_tokens else None,
            status=status,
            incomplete_reason=incomplete_reason,
        )


def opt_int(value: Any) -> int | None:
    if isinstance(value, int):
        return value
    return None


def opt_str(value: Any) -> str | None:
    if isinstance(value, str) and value:
        return value
    return None


def incomplete_reason(original: dict[str, Any]) -> str | None:
    details = original.get("incomplete_details")
    if isinstance(details, dict):
        reason = details.get("reason")
        if isinstance(reason, str) and reason:
            return reason
    return None



from __future__ import annotations

import logging
from typing import Any, Literal

from minibot.core.channels import RenderableResponse
from minibot.shared.assistant_response import coerce_should_answer, payload_to_object


def plain_render(text: str) -> RenderableResponse:
    return RenderableResponse(kind="text", text=text)


def extract_answer(payload: Any, *, logger: logging.Logger) -> tuple[RenderableResponse, bool]:
    payload_obj = payload_to_object(payload)
    if payload_obj is not None:
        answer = payload_obj.get("answer")
        should = payload_obj.get("should_answer_to_user")
        render = render_from_payload(answer, logger=logger)
        should_flag = coerce_should_answer(should)
        if render is not None and should_flag is not None:
            logger.debug(
                "structured output extracted from dict payload",
                extra={"kind": render.kind, "has_answer_object": isinstance(answer, dict)},
            )
            return render, should_flag
        if render is not None and should is None:
            logger.debug(
                "structured output missing should_answer_to_user; defaulting to true",
                extra={"kind": render.kind},
            )
            return render, True
        result = payload_obj.get("result")
        if isinstance(result, str):
            return plain_render(result), True
        timestamp = payload_obj.get("timestamp")
        if isinstance(timestamp, str):
            return plain_render(timestamp), True
        logger.debug(
            "parsed payload looked structured but failed validation",
            extra={
                "parsed_keys": sorted(str(key) for key in payload_obj.keys()),
                "should_type": type(should).__name__,
            },
        )
    if isinstance(payload, str):
        return plain_render(payload), True
    return plain_render(str(payload)), True


def render_from_payload(value: Any, *, logger: logging.Logger) -> RenderableResponse | None:
    if isinstance(value, str):
        logger.debug("structured output answer is legacy string; forcing text kind")
        return plain_render(value)
    if not isinstance(value, dict):
        return None

    content_value = value.get("content")
    if not isinstance(content_value, str):
        legacy_text = value.get("text")
        if isinstance(legacy_text, str):
            content_value = legacy_text

    raw_kind = value.get("kind")
    normalized_kind = normalize_render_kind(raw_kind)
    meta_value = value.get("meta")
    normalized_meta = meta_value if isinstance(meta_value, dict) else {}

    if isinstance(content_value, str) and normalized_kind is not None:
        if not isinstance(meta_value, dict) and meta_value is not None:
            logger.debug(
                "structured output meta is not an object; coercing to empty object",
                extra={"meta_type": type(meta_value).__name__},
            )
        render = RenderableResponse(kind=normalized_kind, text=content_value, meta=normalized_meta)
        logger.debug(
            "structured output answer object normalized",
            extra={
                "kind": render.kind,
                "meta_keys": sorted(render.meta.keys()),
                "source_keys": sorted(str(key) for key in value.keys()),
            },
        )
        if not render.text.strip():
            return None
        return render

    try:
        render = RenderableResponse.model_validate(value)
    except Exception as exc:
        text = content_value
        if isinstance(text, str):
            logger.debug(
                "structured output answer object invalid; using plain text fallback",
                extra={
                    "available_keys": sorted(str(key) for key in value.keys()),
                    "validation_error": str(exc),
                    "raw_kind": raw_kind,
                },
            )
            return plain_render(text)
        return None
    if not render.text.strip():
        return None
    logger.debug(
        "structured output answer object validated",
        extra={
            "kind": render.kind,
            "meta_keys": sorted(render.meta.keys()),
        },
    )
    return render


def normalize_render_kind(value: Any) -> Literal["text", "html", "markdown_v2"] | None:
    if not isinstance(value, str):
        return None
    normalized = value.strip().lower()
    if normalized in {"text", "plain", "plain_text", "plaintext"}:
        return "text"
    if normalized in {"html", "htm"}:
        return "html"
    if normalized in {"markdown_v2", "markdownv2", "markdown", "md"}:
        return "markdown_v2"
    return None



from __future__ import annotations

import json
from typing import Any, Mapping, Sequence

from llm_async.models.tool_call import ToolCall

from minibot.core.agent_runtime import ToolResult
from minibot.llm.services.models import ToolExecutionRecord
from minibot.llm.tools.base import ToolBinding, ToolContext
from minibot.shared.parse_utils import parse_json_maybe_python_object


_MAX_LOG_ARGUMENT_STRING_CHARS = 300
_MAX_LOG_ARGUMENT_COLLECTION_ITEMS = 20
_SENSITIVE_ARGUMENT_KEY_PARTS = (
    "api_key",
    "apikey",
    "token",
    "secret",
    "password",
    "passwd",
    "authorization",
    "cookie",
)


def tool_name_from_call(call: ToolCall) -> str:
    if call.function:
        function_name = call.function.get("name")
        if isinstance(function_name, str) and function_name:
            return function_name
    if call.name:
        return call.name
    return "unknown_tool"


def parse_tool_call(call: ToolCall) -> tuple[str, dict[str, Any]]:
    if call.function:
        func_name = call.function.get("name")
        arguments = call.function.get("arguments")
        if isinstance(arguments, str):
            arguments_payload = arguments.strip()
            if not arguments_payload:
                arguments_dict = {}
            else:
                try:
                    arguments_dict = decode_tool_arguments(arguments_payload)
                except ValueError as exc:
                    preview = arguments_payload.replace("\n", " ")
                    if len(preview) > 220:
                        preview = f"{preview[:220]}..."
                    raise ValueError(
                        f"Tool call arguments must be a valid JSON object. Received arguments preview: {preview}"
                    ) from exc
        else:
            arguments_dict = dict(arguments or {})
    elif call.name:
        func_name = call.name
        arguments_dict = dict(call.input or {})
    else:
        raise ValueError("Tool call missing function metadata")
    if not func_name:
        raise ValueError("Tool call missing name")
    if not isinstance(arguments_dict, dict):
        raise ValueError("Tool call arguments must be an object")
    return func_name, arguments_dict


def decode_tool_arguments(arguments_payload: str) -> dict[str, Any]:
    candidates = [arguments_payload]
    stripped = arguments_payload.strip()
    if stripped.startswith("```"):
        lines = stripped.splitlines()
        if lines and lines[0].startswith("```"):
            lines = lines[1:]
        if lines and lines[-1].strip() == "```":
            lines = lines[:-1]
        fenced = "\n".join(lines).strip()
        if fenced:
            candidates.append(fenced)

    repaired_candidates: list[str] = []
    for candidate in candidates:
        text = candidate.strip()
        if text.startswith("{"):
            missing = text.count("{") - text.count("}")
            if missing > 0:
                repaired_candidates.append(text + ("}" * missing))
    candidates.extend(repaired_candidates)

    for candidate in candidates:
        parsed = parse_json_maybe_python_object(candidate)
        if parsed is None:
            continue
        if isinstance(parsed, dict):
            return dict(parsed)
    raise ValueError("Tool call arguments must be valid JSON")


def stringify_result(result: Any) -> str:
    if isinstance(result, str):
        return result
    if isinstance(result, (list, dict)):
        return json.dumps(result, ensure_ascii=True, default=str)
    return str(result)


def normalize_tool_result(result: Any) -> ToolResult:
    if isinstance(result, ToolResult):
        return result
    return ToolResult(content=result)


def is_sensitive_argument_key(key: str) -> bool:
    normalized = key.strip().lower().replace("-", "_")
    if not normalized:
        return False
    return any(part in normalized for part in _SENSITIVE_ARGUMENT_KEY_PARTS)


def sanitize_tool_arguments_for_log(arguments: Mapping[str, Any]) -> dict[str, Any]:
    sanitized: dict[str, Any] = {}
    for key, value in arguments.items():
        key_text = str(key)
        if is_sensitive_argument_key(key_text):
            sanitized[key_text] = "***"
            continue
        sanitized[key_text] = sanitize_tool_argument_value(value)
    return sanitized


def sanitize_tool_argument_value(value: Any) -> Any:
    if value is None:
        return None
    if isinstance(value, (bool, int, float)):
        return value
    if isinstance(value, str):
        if len(value) <= _MAX_LOG_ARGUMENT_STRING_CHARS:
            return value
        return f"{value[:_MAX_LOG_ARGUMENT_STRING_CHARS]}..."
    if isinstance(value, list):
        capped = value[:_MAX_LOG_ARGUMENT_COLLECTION_ITEMS]
        sanitized_list = [sanitize_tool_argument_value(item) for item in capped]
        if len(value) > _MAX_LOG_ARGUMENT_COLLECTION_ITEMS:
            sanitized_list.append(f"...(+{len(value) - _MAX_LOG_ARGUMENT_COLLECTION_ITEMS} items)")
        return sanitized_list
    if isinstance(value, dict):
        capped_items = list(value.items())[:_MAX_LOG_ARGUMENT_COLLECTION_ITEMS]
        sanitized_dict: dict[str, Any] = {}
        for item_key, item_value in capped_items:
            item_key_text = str(item_key)
            if is_sensitive_argument_key(item_key_text):
                sanitized_dict[item_key_text] = "***"
            else:
                sanitized_dict[item_key_text] = sanitize_tool_argument_value(item_value)
        if len(value) > _MAX_LOG_ARGUMENT_COLLECTION_ITEMS:
            sanitized_dict["..."] = f"+{len(value) - _MAX_LOG_ARGUMENT_COLLECTION_ITEMS} keys"
        return sanitized_dict
    return str(value)


async def execute_tool_calls_for_runtime(
    tool_calls: Sequence[ToolCall],
    tools: Sequence[ToolBinding],
    context: ToolContext,
    *,
    responses_mode: bool,
    logger: Any,
) -> list[ToolExecutionRecord]:
    tool_map = {binding.tool.name: binding for binding in tools}
    records: list[ToolExecutionRecord] = []
    for call in tool_calls:
        call_id = call.id
        if responses_mode and isinstance(call.input, dict):
            input_call_id = call.input.get("call_id")
            if isinstance(input_call_id, str) and input_call_id:
                call_id = input_call_id
        tool_name = tool_name_from_call(call)
        try:
            tool_name, arguments = parse_tool_call(call)
            binding = tool_map.get(tool_name)
            if not binding:
                raise ValueError(f"tool {tool_name} is not registered")
            logger.debug(
                "executing tool",
                extra={
                    "tool": tool_name,
                    "call_id": call_id,
                    "owner_id": context.owner_id,
                    "argument_keys": sorted(arguments.keys()),
                    "arguments": sanitize_tool_arguments_for_log(arguments),
                },
            )
            raw_result = await binding.handler(arguments, context)
            result = normalize_tool_result(raw_result)
            logger.debug(
                "tool execution completed",
                extra={
                    "tool": tool_name,
                    "call_id": call_id,
                    "owner_id": context.owner_id,
                },
            )
        except Exception as exc:
            error_code = "tool_execution_failed"
            if isinstance(exc, ValueError) and "arguments" in str(exc).lower():
                error_code = "invalid_tool_arguments"
            logger.exception(
                "tool execution failed",
                extra={
                    "tool": tool_name,
                    "owner_id": context.owner_id,
                },
            )
            result = ToolResult(
                content={
                    "ok": False,
                    "tool": tool_name,
                    "error_code": error_code,
                    "error": str(exc),
                }
            )
        if responses_mode:
            payload = {
                "type": "function_call_output",
                "call_id": call_id,
                "output": stringify_result(result.content),
            }
        else:
            payload = {
                "role": "tool",
                "tool_call_id": call_id,
                "name": tool_name,
                "content": stringify_result(result.content),
            }
        records.append(
            ToolExecutionRecord(
                tool_name=tool_name,
                call_id=call_id,
                message_payload=payload,
                result=result,
            )
        )
    return records


async def execute_tool_calls(
    tool_calls: Sequence[ToolCall],
    tools: Sequence[ToolBinding],
    context: ToolContext,
    *,
    responses_mode: bool,
    logger: Any,
) -> list[dict[str, Any]]:
    records = await execute_tool_calls_for_runtime(
        tool_calls,
        tools,
        context,
        responses_mode=responses_mode,
        logger=logger,
    )
    return [record.message_payload for record in records]



from __future__ import annotations

import json
import logging
import re
from dataclasses import dataclass
from typing import Any, Protocol, Sequence

from minibot.core.agent_runtime import AgentState
from minibot.llm.tools.base import ToolBinding, ToolContext
from minibot.shared.path_utils import normalize_path_separators


@dataclass(frozen=True)
class GuardrailDecision:
    requires_retry: bool
    retry_system_prompt_suffix: str | None = None
    resolved_render_text: str | None = None
    suggested_tool: str | None = None
    suggested_path: str | None = None
    tokens_used: int = 0


class ToolUseGuardrail(Protocol):
    async def apply(
        self,
        *,
        session_id: str,
        user_text: str,
        tool_context: ToolContext,
        state: AgentState,
        system_prompt: str,
        prompt_cache_key: str | None,
    ) -> GuardrailDecision: ...


class NoopToolUseGuardrail:
    async def apply(self, **_: Any) -> GuardrailDecision:
        return GuardrailDecision(requires_retry=False)


class LLMClassifierToolUseGuardrail:
    def __init__(
        self,
        *,
        llm_client: Any,
        tools: Sequence[ToolBinding],
    ) -> None:
        self._llm_client = llm_client
        self._tools = list(tools)
        self._logger = logging.getLogger("minibot.tool_use_guardrail")

    async def apply(
        self,
        *,
        session_id: str,
        user_text: str,
        tool_context: ToolContext,
        state: AgentState,
        system_prompt: str,
        prompt_cache_key: str | None,
    ) -> GuardrailDecision:
        if not self._tools:
            return GuardrailDecision(requires_retry=False)

        tool_names = [binding.tool.name for binding in self._tools]
        classifier_prompt = (
            "Decide whether the user's request requires executing at least one tool before answering. "
            "Use the available tool names exactly as given. "
            "Return structured output only.\n\n"
            f"Available tools: {', '.join(tool_names)}\n"
            f"User request:\n{user_text}"
        )
        try:
            history = [m for m in state.messages if m.role in {"user", "assistant", "tool"}]
            generation = await self._llm_client.generate(
                history,
                classifier_prompt,
                user_content=None,
                tools=[],
                tool_context=None,
                response_schema=self._schema(),
                prompt_cache_key=f"{prompt_cache_key}:tool-requirement" if prompt_cache_key else None,
                previous_response_id=None,
                system_prompt_override="You are a strict tool-routing classifier.",
            )
            raw_tokens = getattr(generation, "total_tokens", None)
            tokens = raw_tokens if isinstance(raw_tokens, int) and raw_tokens > 0 else 0
            payload = generation.payload
            payload_obj: dict[str, Any] | None = None
            if isinstance(payload, dict):
                payload_obj = payload
            elif isinstance(payload, str) and payload.strip():
                parsed = json.loads(payload.strip())
                if isinstance(parsed, dict):
                    payload_obj = parsed
            if not payload_obj or not bool(payload_obj.get("requires_tools", False)):
                return GuardrailDecision(requires_retry=False, tokens_used=tokens)

            raw_tool = payload_obj.get("suggested_tool")
            suggested_tool = raw_tool if isinstance(raw_tool, str) and raw_tool in tool_names else None
            raw_path = payload_obj.get("path")
            suggested_path = raw_path.strip() if isinstance(raw_path, str) and raw_path.strip() else None

            direct_result = await self._try_direct_delete(
                user_text=user_text,
                tool_context=tool_context,
                suggested_tool=suggested_tool,
                suggested_path=suggested_path,
            )
            if direct_result is not None:
                return GuardrailDecision(
                    requires_retry=False,
                    resolved_render_text=direct_result,
                    tokens_used=tokens,
                )

            self._logger.debug(
                "tool requirement classifier requires retry",
                extra={"suggested_tool": suggested_tool, "suggested_path": suggested_path},
            )
            return GuardrailDecision(
                requires_retry=True,
                retry_system_prompt_suffix=(
                    "Tool policy reminder: this request requires using tools before final answer. "
                    "Call the relevant tool now, then provide the final answer from tool output. "
                    "Do not answer with intent statements like 'I will check'."
                ),
                suggested_tool=suggested_tool,
                suggested_path=suggested_path,
                tokens_used=tokens,
            )
        except Exception:
            self._logger.exception("tool requirement classifier failed; skipping guardrail")
            return GuardrailDecision(requires_retry=False)

    async def _try_direct_delete(
        self,
        *,
        user_text: str,
        tool_context: ToolContext,
        suggested_tool: str | None,
        suggested_path: str | None,
    ) -> str | None:
        if suggested_tool not in {"delete_file", "filesystem"}:
            return None
        binding = next((item for item in self._tools if item.tool.name == "filesystem"), None)
        if binding is None:
            return None
        candidates = _extract_delete_path_candidates(user_text)
        if suggested_path is not None and suggested_path not in candidates:
            candidates.insert(0, suggested_path)
        if not candidates:
            return None
        for candidate in candidates:
            try:
                raw_result = await binding.handler({"action": "delete", "path": candidate}, tool_context)
                if not isinstance(raw_result, dict):
                    continue
                if int(raw_result.get("deleted_count") or 0) > 0:
                    return str(raw_result.get("message") or f"Deleted file successfully: {candidate}")
            except Exception:
                self._logger.exception("direct filesystem delete failed", extra={"path": candidate})
        return None

    @staticmethod
    def _schema() -> dict[str, Any]:
        return {
            "type": "object",
            "properties": {
                "requires_tools": {"type": "boolean"},
                "suggested_tool": {"type": ["string", "null"]},
                "path": {"type": ["string", "null"]},
            },
            "required": ["requires_tools", "suggested_tool", "path"],
            "additionalProperties": False,
        }


def _extract_delete_path_candidates(text: str) -> list[str]:
    candidates: list[str] = []

    def _normalize(value: str) -> str | None:
        value = value.strip().strip('"').strip("'")
        if not value:
            return None
        if value.startswith("./"):
            value = value[2:]
        return normalize_path_separators(value)

    for item in re.findall(r"['\"]([^'\"]+)['\"]", text):
        n = _normalize(item)
        if n and n not in candidates:
            candidates.append(n)

    for item in re.findall(r"(?:\.?/?[a-zA-Z0-9_-]+/)+[a-zA-Z0-9_.-]+\.[a-zA-Z0-9]{1,10}", text):
        n = _normalize(item)
        if n and n not in candidates:
            candidates.append(n)

    m = re.search(r"([a-zA-Z0-9_.-]+\.[a-zA-Z0-9]{1,10})", text)
    if m:
        n = _normalize(m.group(1))
        if n and n not in candidates:
            candidates.append(n)

    return candidates



from __future__ import annotations

from functools import lru_cache
from importlib.resources import files


@lru_cache(maxsize=None)
def load_tool_description(name: str) -> str:
    package = files("minibot.llm.tools.descriptions")
    resource = package.joinpath(f"{name}.txt")
    try:
        text = resource.read_text(encoding="utf-8").strip()
    except (FileNotFoundError, TypeError) as exc:
        raise FileNotFoundError(f"Missing tool description file: descriptions/{name}.txt") from exc
    if not text:
        raise ValueError(f"Empty tool description file: descriptions/{name}.txt")
    return text



from __future__ import annotations

from dataclasses import dataclass
import logging
from typing import Any, Sequence, cast

from minibot.app.agent_runtime import AgentRuntime
from minibot.app.delegation_trace import count_tool_messages, extract_delegation_trace
from minibot.app.response_parser import extract_answer, plain_render
from minibot.app.tool_use_guardrail import ToolUseGuardrail
from minibot.core.agent_runtime import AgentMessage, AgentState, MessagePart, MessageRole
from minibot.llm.provider_factory import LLMClient
from minibot.llm.tools.base import ToolContext

from minibot.app.handlers.services.session_state_service import SessionStateService


@dataclass(frozen=True)
class AgentRuntimeResult:
    render: Any
    should_reply: bool
    response_id: str | None
    agent_trace: list[dict[str, Any]]
    delegation_fallback_used: bool
    tokens_used: int


class RuntimeOrchestrationService:
    def __init__(
        self,
        *,
        runtime: AgentRuntime,
        llm_client: LLMClient,
        guardrail: ToolUseGuardrail,
        session_state: SessionStateService,
        logger: logging.Logger,
    ) -> None:
        self._runtime = runtime
        self._llm_client = llm_client
        self._guardrail = guardrail
        self._session_state = session_state
        self._logger = logger

    async def run_with_agent_runtime(
        self,
        *,
        session_id: str,
        history: list[Any],
        model_text: str,
        model_user_content: str | list[dict[str, Any]] | None,
        system_prompt: str,
        tool_context: ToolContext,
        prompt_cache_key: str | None,
        previous_response_id: str | None,
        chat_id: int | None,
        channel: str | None,
        response_schema: dict[str, Any],
    ) -> AgentRuntimeResult:
        tokens_used = 0
        trace_result = None

        state = self._build_agent_state(
            history=history,
            user_text=model_text,
            user_content=model_user_content,
            system_prompt=system_prompt,
        )
        generation = await self._runtime.run(
            state=state,
            tool_context=tool_context,
            response_schema=response_schema,
            prompt_cache_key=prompt_cache_key,
            initial_previous_response_id=previous_response_id,
        )
        tokens_used += self._session_state.track_tokens(session_id, getattr(generation, "total_tokens", None))
        tool_messages_count = count_tool_messages(generation.state)
        trace_result = extract_delegation_trace(generation.state)
        delegation_unresolved = trace_result.unresolved

        guardrail = await self._guardrail.apply(
            session_id=session_id,
            user_text=model_text,
            tool_context=tool_context,
            state=generation.state,
            system_prompt=system_prompt,
            prompt_cache_key=prompt_cache_key,
        )
        tokens_used += guardrail.tokens_used
        self._session_state.track_tokens(session_id, guardrail.tokens_used)

        if guardrail.resolved_render_text is not None:
            return AgentRuntimeResult(
                render=plain_render(guardrail.resolved_render_text),
                should_reply=True,
                response_id=generation.response_id,
                agent_trace=trace_result.trace,
                delegation_fallback_used=trace_result.fallback_used,
                tokens_used=tokens_used,
            )

        if guardrail.requires_retry and tool_messages_count == 0:
            retry_system_prompt = system_prompt
            if guardrail.retry_system_prompt_suffix:
                retry_system_prompt = f"{system_prompt}\n\n{guardrail.retry_system_prompt_suffix}"
            retry_state = self._build_agent_state(
                history=history,
                user_text=model_text,
                user_content=model_user_content,
                system_prompt=retry_system_prompt,
            )
            generation = await self._runtime.run(
                state=retry_state,
                tool_context=tool_context,
                response_schema=response_schema,
                prompt_cache_key=prompt_cache_key,
                initial_previous_response_id=generation.response_id,
            )
            tokens_used += self._session_state.track_tokens(session_id, getattr(generation, "total_tokens", None))
            trace_result = extract_delegation_trace(generation.state)
            delegation_unresolved = trace_result.unresolved
            render, should_reply = extract_answer(generation.payload, logger=self._logger)
            if count_tool_messages(generation.state) == 0:
                return AgentRuntimeResult(
                    render=plain_render(
                        "I could not verify or execute that action with tools in this attempt. "
                        "Please try again, or ask me to run a specific tool."
                    ),
                    should_reply=True,
                    response_id=generation.response_id,
                    agent_trace=trace_result.trace,
                    delegation_fallback_used=trace_result.fallback_used,
                    tokens_used=tokens_used,
                )
        else:
            render, should_reply = extract_answer(generation.payload, logger=self._logger)

        if delegation_unresolved:
            retry_state = self._build_agent_state(
                history=history,
                user_text=model_text,
                user_content=model_user_content,
                system_prompt=(
                    f"{system_prompt}\n\n"
                    "Delegation policy reminder: If invoke_agent result has should_answer_to_user=false "
                    "or result_status other than success, you must resolve it in this turn. "
                    "Do one additional concrete tool call or return an explicit failure message to user "
                    "with should_answer_to_user=true."
                ),
            )
            generation = await self._runtime.run(
                state=retry_state,
                tool_context=tool_context,
                response_schema=response_schema,
                prompt_cache_key=prompt_cache_key,
                initial_previous_response_id=generation.response_id,
            )
            tokens_used += self._session_state.track_tokens(session_id, getattr(generation, "total_tokens", None))
            trace_result = extract_delegation_trace(generation.state)
            delegation_unresolved = trace_result.unresolved
            render, should_reply = extract_answer(generation.payload, logger=self._logger)
            if delegation_unresolved:
                self._logger.warning(
                    "delegation unresolved after bounded retry; returning explicit failure",
                    extra={"chat_id": chat_id, "channel": channel},
                )
                render = plain_render(
                    "I could not complete that delegated action reliably in this attempt. "
                    "Please retry, or ask me to run a specific tool step-by-step."
                )
                should_reply = True

        assert trace_result is not None
        return AgentRuntimeResult(
            render=render,
            should_reply=should_reply,
            response_id=generation.response_id,
            agent_trace=trace_result.trace,
            delegation_fallback_used=trace_result.fallback_used,
            tokens_used=tokens_used,
        )

    @staticmethod
    def _build_agent_state(
        history: Sequence[Any],
        user_text: str,
        user_content: str | list[dict[str, Any]] | None,
        system_prompt: str,
    ) -> AgentState:
        messages: list[AgentMessage] = [
            AgentMessage(role="system", content=[MessagePart(type="text", text=system_prompt)])
        ]
        for entry in history:
            role = str(getattr(entry, "role", "user"))
            if role not in {"system", "user", "assistant"}:
                continue
            content = getattr(entry, "content", "")
            messages.append(
                AgentMessage(
                    role=cast(MessageRole, role),
                    content=[MessagePart(type="text", text=str(content))],
                )
            )

        if user_content is None:
            messages.append(AgentMessage(role="user", content=[MessagePart(type="text", text=user_text)]))
        elif isinstance(user_content, str):
            messages.append(AgentMessage(role="user", content=[MessagePart(type="text", text=user_content)]))
        else:
            messages.append(
                AgentMessage(
                    role="user",
                    content=[MessagePart(type="text", text=user_text)],
                    raw_content=user_content,
                )
            )
        return AgentState(messages=messages)



from __future__ import annotations

from dataclasses import dataclass
from typing import Any

from minibot.core.agent_runtime import ToolResult


@dataclass
class LLMGeneration:
    payload: Any
    response_id: str | None = None
    total_tokens: int | None = None
    input_tokens: int | None = None
    output_tokens: int | None = None
    cached_input_tokens: int | None = None
    reasoning_output_tokens: int | None = None
    status: str | None = None
    incomplete_reason: str | None = None


@dataclass
class LLMCompletionStep:
    message: Any
    response_id: str | None
    total_tokens: int | None = None


@dataclass
class ToolExecutionRecord:
    tool_name: str
    call_id: str
    message_payload: dict[str, Any]
    result: ToolResult


@dataclass
class LLMCompaction:
    response_id: str
    output: list[dict[str, Any]]
    total_tokens: int | None = None


@dataclass(frozen=True)
class UsageSnapshot:
    total_tokens: int | None = None
    input_tokens: int | None = None
    output_tokens: int | None = None
    cached_input_tokens: int | None = None
    reasoning_output_tokens: int | None = None
    status: str | None = None
    incomplete_reason: str | None = None
